# -*- coding: utf-8 -*-

"""
@author: Fraser M. Callaghan

Input / Output functions
"""


import os
import sys
import numpy as np
import unicodedata as ud
import pickle
import string
import json
from datetime import date, datetime, timedelta
import codecs
import csv
import shutil
import vtk
from vtk.util import numpy_support # type: ignore
try:
    import xml.etree.cElementTree as ET
except ImportError:
    import xml.etree.ElementTree as ET
from typing import List, Dict, Union, Optional, Tuple
import tarfile

def remove_symbols(text: str) -> str:
    """
    Remove symbols from the given text.

    Args:
        text (str): The input text to process.

    Returns:
        str: The text with symbols removed.
    """
    return ''.join(c for c in text if ud.category(c) != 'So')

def remove_not_allowed(text: str, allowed_characters: str = string.printable + 'ßöüäéèà') -> str:
    """
    Remove characters that are not in the allowed set.

    Args:
        text (str): The input text to process.
        allowed_characters (str): The set of characters to keep (default is all printable characters and some special (DE, FR) characters).

    Returns:
        str: The text with disallowed characters removed.
    """
    return "".join(c for c in text if c in allowed_characters)

# =========================================================================
##          Utilities
# =========================================================================
def countFilesInDir(dirName: str) -> int:
    """
    Count the number of files in a directory.

    Args:
        dirName (str): The path to the directory.

    Returns:
        int: The number of files in the directory.
    """
    files = []
    if os.path.isdir(dirName):
        for _, _, filenames in os.walk(dirName):  
            files.extend(filenames)
    return len(files)

def getAllFilesUnderDir(dirName: str) -> List[str]:
    """
    Get all file paths under a directory.

    Args:
        dirName (str): The path to the directory.

    Returns:
        List[str]: A list of full file paths.
    """
    files = []
    if os.path.isdir(dirName):
        for path, _, filenames in os.walk(dirName):
            for ifile in filenames:
                files.append(os.path.join(path, ifile))
    return files

def parseFileToTagsDictionary(fileName: str, equator: str = '=', commentor: str = '#', SPACE_REPLACE: bool = True) -> Dict[str, str]:
    """
    Parse a file to a dictionary of tags.

    Args:
        fileName (str): The name of the file to parse.
        equator (str, optional): The character used to separate keys and values. Defaults to '='.
        commentor (str, optional): The character used to denote comments. Defaults to '#'.
        SPACE_REPLACE (bool, optional): Whether to replace spaces in keys with underscores. Defaults to True.

    Returns:
        dict: A dictionary containing the parsed tags and values.

    Notes:
        - If the file ends with '.json', it will be parsed as a JSON file instead.
        - Comments at the beginning of the file are stored under the 'header' key.
        - Subsequent comments are stored as 'comment0', 'comment1', etc.
        - Empty lines and lines without the equator character are ignored.
        - Keys and values are stored as strings.
    """
    if fileName.endswith('json'):
        return parseJsonToDictionary(fileName)
    myDict, commentNumber = {'header': ''}, 0
    HEADER_FLAG = True
    try:
        with open(fileName, 'r') as f:
            for lineIn in f:
                if (len(lineIn) > 1):
                    if (lineIn[0] == commentor):  #  comments
                        if HEADER_FLAG:
                            myDict['header'] = myDict['header'] + lineIn
                        else:
                            myDict['comment%d'%(commentNumber)] = lineIn[1:]
                            commentNumber += 1
                    else:
                        HEADER_FLAG = False
                        myEquals = lineIn.find(equator)             # Find = sign
                        if myEquals < 0:
                            continue
                        else:
                            myKey = lineIn[:myEquals].strip()
                            if SPACE_REPLACE:
                                myKey = myKey.replace(' ', '_')
                            myValue = lineIn[myEquals+1:].strip()
                            myValue = myValue.rstrip('\n')
                            myDict[myKey] = myValue
    except IOError:
        raise
    if myDict['header'] == '':
        myDict.pop('header')
    return myDict


def writeDictionaryToFile(fileName: str, dictToWrite: Dict[str, str], equator: str = '=', commentor: str = '#', WRITE_COMMENTS: bool = True) -> str:
    """
    Write a dictionary to a file.

    Args:
        fileName (str): The name of the file to write to (if json then calls writeDictionaryToJSON - json.dump with Numpy encoder).
        dictToWrite (dict): The dictionary to write to the file.
        equator (str, optional): The character used to separate keys and values. Defaults to '='.
        commentor (str, optional): The character used to denote comments. Defaults to '#'.
        WRITE_COMMENTS (bool, optional): Whether to write comments to the file. Defaults to True.

    Returns:
        str: The name of the file written to.
    """
    if fileName.endswith('json'):
        return writeDictionaryToJSON(fileName, dictToWrite)
    with open(fileName, 'w') as fid:
        try:
            fid.write('%s'%(dictToWrite['header']))
        except KeyError:
            pass # no header
        for iKey in sorted(dictToWrite.keys()):
            if 'comment' in iKey.lower():
                if WRITE_COMMENTS:
                    fid.write('%s %s\n'%(commentor, remove_not_allowed(str(dictToWrite[iKey]))))
            elif iKey.lower() == 'header':
                continue
            else:
                fid.write('%s %s %s\n'%(iKey, equator, remove_not_allowed(str(dictToWrite[iKey]))))
        return fileName


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.bool_):
            return bool(obj)
        if isinstance(obj, (np.floating, np.complexfloating)):
            return float(obj)
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, np.string_):
            return str(obj)
        if isinstance(obj, (datetime, date)):
            return obj.isoformat()
        if isinstance(obj, timedelta):
            return str(obj)
        return super(NumpyEncoder, self).default(obj)


def writeDictionaryToJSON(fileName: str, dictToWrite: Dict[str, str]) -> str:
    with open(fileName, 'w') as fp:
        json.dump(dictToWrite, fp, indent=4, sort_keys=True, cls=NumpyEncoder, ensure_ascii=False)
    return fileName


def parseJsonToDictionary(fileName: str) -> Dict[str, str]:
    with open(fileName, 'r') as fid:
        myDict = json.load(fid)
    return myDict


def strListToFloatList(strList: str) -> List[float]:
    """
    Convert a string representation of a list to a list of floats.

    Args:
        strList (str): A string representation of a list of numbers.

    Returns:
        List[float]: A list of float values.
    """
    strList = strList.replace("[","")
    strList = strList.replace("]","")
    return [float(i) for i in strList.split(",")]

def xmlIndent(elem: ET.Element, level: int = 0) -> None:
    """
    Indent an XML element for pretty printing.

    Args:
        elem (ET.Element): The XML element to indent.
        level (int, optional): The current indentation level. Defaults to 0.
    """
    i = "\n" + level*"  "
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + "  "
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for elem in elem:
            xmlIndent(elem, level+1)
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i

def __buildFileName(prefix: str, idNumber: int, extn: str, number_of_digits: int = 6) -> str:
    """
    Build a file name from a prefix, ID number, and extension.

    Args:
        prefix (str): The prefix of the file name.
        idNumber (int): The ID number of the file.
        extn (str): The extension of the file.

    Returns:
        str: The file name.
    """
    ids = f'{idNumber:0{number_of_digits}d}'
    if extn[0] != '.':
        extn = '.' + extn
    fileName = prefix + '_' + ids + extn
    return fileName

def findAndReplaceTextInFile(fileIn: str, txtFind: str, txtReplace: str) -> str:
    """
    Find and replace text in a file.

    Args:
        fileIn (str): The input file path.
        txtFind (str): The text to find.
        txtReplace (str): The text to replace with.

    Returns:
        str: The path of the modified file.
    """
    with open(fileIn, 'r') as fid:
        fileData = fid.read()
    fileData = fileData.replace(txtFind, txtReplace)
    with open(fileIn, 'w') as fid:
        fid.write(fileData)
    return fileIn

def readFileToListOfLines(fileName: str, commentSymbol: str = '#') -> List[List[str]]:
    """
    Read a file and return a list of each line.

    Args:
        fileName (str): The path of the file to read.
        commentSymbol (str, optional): The symbol used to denote comments. Defaults to '#'.

    Returns:
        List[List[str]]: A list of lines from the file.

    Notes:
        - Will split on "," if present
        - Will skip starting with #
    """
    with open(fileName, 'r') as fid:
        lines = fid.readlines()
    lines = [l.strip('\n') for l in lines]
    lines = [l for l in lines if len(l) > 0]
    lines = [l for l in lines if l[0]!=commentSymbol]
    lines = [l.split(',') for l in lines]
    return lines
 
def appendFiles(listOfFiles: List[str], carriageReturn: str = '') -> str:
    """
    Append the contents of multiple text files into a single string.

    Args:
        listOfFiles (list): A list of file names to append.
        carriageReturn (str, optional): The string to append after each file's content. Defaults to an empty string.

    Returns:
        str: A single string containing the concatenated contents of all files.
    """
    allText = ''
    for iFile in listOfFiles:
        with open(iFile, 'r') as fid:
            fileData = fid.read()
            allText = allText + fileData + carriageReturn
    return allText

def pickleData(data, pickleFileName: str) -> str:
    """
    Pickle data to a file.

    Args:
        data: The data to pickle.
        pickleFileName (str): The name of the file to pickle to.

    Returns:
        str: The name of the file written to.
    """
    with open(pickleFileName, 'wb') as fid:
        pickle.dump(data, fid, protocol=2)
    return pickleFileName

def unpickleData(pickleFileName: str):
    """
    Unpickle data from a file.

    Args:
        pickleFileName (str): The name of the file to unpickle.

    Returns:
        The unpickled data.
    """
    opts = {}
    if sys.version_info[0] == 3:
        opts = {"encoding":'latin1'}
    with open(pickleFileName, 'rb') as fid:
        return pickle.load(fid, **opts)


def tarGZLocal_AndMove(dirToTarZip: str, tarGZFileName: str, localLocation: str, remoteLocation: Optional[str] = None) -> None:
    """
    Compress and move a directory to a local location and optionally to a remote location.
    Uses Python's tarfile module for cross-platform compatibility.

    Args:
        dirToTarZip (str): The directory to compress and zip.
        tarGZFileName (str): The name of the compressed file.
        localLocation (str): The local location to move the compressed file.
        remoteLocation (str, optional): The remote location to move the compressed file. Defaults to None.
    """
    # Get the parent directory and target directory name
    parent_dir = os.path.dirname(dirToTarZip)
    target_dir = os.path.basename(dirToTarZip)
    output_file = os.path.join(localLocation, tarGZFileName)
    
    # Create tar.gz archive
    with tarfile.open(output_file, "w:gz") as tar:
        # Change to parent directory to get relative paths
        original_dir = os.getcwd()
        os.chdir(parent_dir)
        try:
            tar.add(target_dir)
        finally:
            os.chdir(original_dir)
    
    if remoteLocation is not None:
        shutil.copy(output_file, remoteLocation)
        os.unlink(output_file)


# =========================================================================
# =========================================================================
def addSuffixToName(fileName: str, suffix: str) -> str:
    """
    Add a suffix to the end of a file name (retains extn).

    Args:
        fileName (str): The name of the file to modify.
        suffix (str): The suffix to add to the file name.

    Returns:
        str: The modified file name.
    """
    fileName, fileExt = os.path.splitext(fileName)
    fSplit = fileName.split('_')
    number = fSplit[-1]
    return fileName.replace('_' + number, suffix + '_' + number) + fileExt

def checkIfExtnPresent(fileName: str, extn: str) -> str:
    """
    Check if the extension of a file name is present and add it if it is not.

    Args:
        fileName (str): The name of the file to check.
        extn (str): The extension to check for.

    Returns:
        str: The modified file name with the extension added if it was not present.
    """
    if (extn[0] == '.'):
        extn = extn[1:]

    le = len(extn)
    if (fileName[-le:] != extn):
        fileName = fileName + '.' + extn
    return fileName


# =========================================================================
##          CSV interface
# =========================================================================

def readDataCSV(csvFile: str, nHeaderLines: int = 0, quotechar: str = '"', decimalChar: str = ".", delimeterToUse: Optional[str] = None) -> Tuple[List[str], List[List[str]]]:
    """
    Read data from a CSV file.

    Args:
        csvFile (str): The path of the CSV file to read.
        nHeaderLines (int, optional): The number of header lines to skip. Defaults to 0.
        quotechar (str, optional): The character used to quote fields. Defaults to '"'.
        decimalChar (str, optional): The character used to denote decimal points. Defaults to '.'.
        delimeterToUse (str, optional): The delimiter to use. If not specified, it will be detected automatically.

    Returns:
        Tuple[List[str], List[List[str]]]: A tuple containing the headers and the data.

    Notes:
        - Assumes Headers: then cols of floats
        - Try to automatically determine detectDelimiter
        - Read file (skip header lines)
        Return last headers row as list of headers (if so) & list of lists of each row
        Can pass output to np.array() then have columns of data
            e.g. : data = np.array(data, dtype=float)
    """
    def __fixDecimal(irow: List[str]) -> List[str]:
        if decimalChar != ".":
            rowOut = []
            for ii in irow:
                i2 = ii.replace(decimalChar, ".")
                try:
                    float(i2)
                    rowOut.append(i2)
                except ValueError:
                    rowOut.append(ii)
            return rowOut
        return irow
    
    def doReading(delimeterToUse_: str, headers: List[str], infile, nHeaderLines: int, quotechar: str) -> Tuple[List[str], List[List[str]]]:
        reader = csv.reader(infile, delimiter=delimeterToUse_, quotechar=quotechar)
        # ASSUME ALWAYS SAME ROW FORMAT
        rowsList = []
        rowCount = 0
        for row in reader:
            rowCount += 1
            if rowCount <= nHeaderLines:
                headers = row
                continue
            row = __fixDecimal(row)
            rowsList.append(row)
        return headers, rowsList
    if delimeterToUse is None:
        try:
            delimeterToUse = detectDelimiter(csvFile, nHeaderLines)
        except UnicodeDecodeError:
            delimeterToUse = detectDelimiter(csvFile, nHeaderLines, 'utf16')
    headers = []
    try:
        with open(csvFile, 'rU') as infile:
            headers, rowsList = doReading(delimeterToUse, headers, infile, nHeaderLines, quotechar)
    except (csv.Error, UnicodeDecodeError):
        with codecs.open(csvFile, 'rU', 'utf-16') as infile:
            headers, rowsList = doReading(delimeterToUse, headers, infile, nHeaderLines, quotechar)
    return headers, rowsList


def writeCSVFile(data: List[List[str]], header: List[str], csvFile: str, FIX_NAN: bool = False) -> str:
    """
    Write data to a CSV file.

    Args:
        data (List[List[str]]): The data to write to the CSV file.
        header (List[str]): The header for the CSV file.
        csvFile (str): The path of the CSV file to write to.
        FIX_NAN (bool, optional): Whether to fix NaN values. Defaults to False.

    Returns:
        str: The path of the CSV file written to.
    """
    with open(csvFile, 'w') as fout:
        csvWriter = csv.writer(fout, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
        #first write column headers
        if header is not None:
            csvWriter.writerow(header)
        for iRow in data:
            if FIX_NAN:
                iRow = ['' if i=='nan' else i for i in iRow]
            csvWriter.writerow(iRow)
    return csvFile


def detectDelimiter(csvFile: str, nHeader: int = 0, encoding: Optional[str] = None) -> str:
    """
    Detect the delimiter used in a CSV file.

    Args:
        csvFile (str): The path of the CSV file to detect the delimiter of.
        nHeader (int, optional): The number of header lines to skip. Defaults to 0.
        encoding (str, optional): The encoding of the file. Defaults to None.

    Returns:
        str: The detected delimiter.
    """
    with open(csvFile, 'r', encoding=encoding) as myCsvfile:
        c0 = 0
        for row in myCsvfile:
            while c0 <= nHeader:
                c0 += 1
                continue
            if row.find(",")!=-1:
                return ","
            if row.find(";")!=-1:
                return ";"
            if row.find("\t")!=-1:
                return "\t"
    return ","


# =========================================================================
##          PLY interface
# =========================================================================

def writePlyFile(fullfileName: str, xyzPoints: List[List[float]], uvwPoints: Optional[List[List[float]]] = None, comment: List[str] = ["Written via fIO module"], header: bool = True) -> str:
    """
    Write points and norms to PLY file
    :param fullfileName:
    :param xyzPoints:
    :param uvwPoints:
    :param comment:
    :param header:
    :return:
    """
    # FORMAT EXAMPLE :
    # ply
    # format ascii 1.0           { ascii/binary, format version number }
    # comment made by Greg Turk  { comments keyword specified, like all lines }
    # comment this file is a cube
    # element vertex 8           { define "vertex" element, 8 of them in file }
    # property float x           { vertex contains float "x" coordinate }
    # property float y           { y coordinate is also a vertex property }
    # property float z           { z coordinate, too }
    # element face 6             { there are 6 "face" elements in the file }
    # property list uchar int vertex_index { "vertex_indices" is a list of ints }
    # end_header                 { delimits the end of the header }
    # 0 0 0                      { start of vertex list }
    # 0 0 1
    # 0 1 1
    # 0 1 0
    # 1 0 0
    # 1 0 1
    # 1 1 1
    # 1 1 0
    # 4 0 1 2 3                  { start of face list }
    # 4 7 6 5 4
    # 4 0 4 5 1
    # 4 1 5 6 2
    # 4 2 6 7 3
    # 4 3 7 4 0
    # -------------------
    nPts = len(xyzPoints)
    with open(fullfileName, 'w') as fid:
        if header:
            fid.write('ply\n')
            fid.write('format ascii 1.0\n')
            for icomment in comment:
                fid.write('comment ' + icomment + '\n')
            fid.write('element vertex %d\n' % (nPts))
            fid.write('property float x\n')
            fid.write('property float y\n')
            fid.write('property float z\n')
            if uvwPoints is not None:
                fid.write('property float nx\n')
                fid.write('property float ny\n')
                fid.write('property float nz\n')
            fid.write('end_header\n')

        if uvwPoints is not None:
            for k in range(nPts):
                fid.write('%.7f %.7f %.7f %.7f %.7f %.7f\n' % \
                          (xyzPoints[k][0], xyzPoints[k][1], xyzPoints[k][2], \
                           uvwPoints[k][0], uvwPoints[k][1], uvwPoints[k][2]))
        else:
            for k in range(nPts):
                fid.write('%.7f %.7f %.7f\n' % \
                          (xyzPoints[k][0], xyzPoints[k][1], xyzPoints[k][2]))
        #
    #
    return fullfileName



# =========================================================================
## vtkIO and PVD interface
# =========================================================================
def writeVTKFile(data: vtk.vtkDataObject, fileName: str, STL_ASCII: bool = False) -> str:
    """
    Writes a VTK object to a file. Uses the extension to determine the type of file to write.

    Args:
        data (vtk.vtkDataObject): The VTK object to write to a file.
        fileName (str): The full file name to write to.
        STL_ASCII (bool, optional): Set to True to write STL file in ASCII [default False -> write binary] (only used for STL).

    Returns:
        str: The full file name.
    """
    writer = None
    fileName = str(fileName)
    if fileName.endswith('vtp'):
        writer = vtk.vtkXMLPolyDataWriter()
    elif fileName.endswith('vts'):
        writer = vtk.vtkXMLStructuredGridWriter()
    elif fileName.endswith('vtu'):
        writer = vtk.vtkXMLUnstructuredGridWriter()
    elif fileName.endswith('stl'):
        writer = vtk.vtkSTLWriter()
        if STL_ASCII:
            writer.SetFileTypeToASCII()
        else:
            writer.SetFileTypeToBinary()
    elif fileName.endswith('vti'):
        writer = vtk.vtkXMLImageDataWriter()
        writer.SetDataModeToBinary()
    elif fileName.endswith('mhd'):
        writer = vtk.vtkMetaImageWriter()
    elif fileName.endswith('mha'):
        writer = vtk.vtkMetaImageWriter()
    elif fileName.endswith('nii'):
        return writeNifti(data, fileName)
    ##
    if not writer:
        raise IOError('Extension not recognised (%s)'%(fileName))
    writer.SetFileName(fileName)
    writer.SetInputData(data)
    writer.Write()
    return fileName


def readVTKFile(fileName: str) -> vtk.vtkDataObject:
    """
    Read a VTK file - the extension is used to determine the type of file to read.

    Args:
        fileName (str): The path of the VTK file to read.

    Returns:
        vtk.vtkDataObject: The VTK data object.
    """
    if not os.path.isfile(fileName):
        raise IOError('## ERROR: %s file not found'%(fileName))
    if fileName.endswith('vtp'):
        reader = vtk.vtkXMLPolyDataReader()
    elif fileName.endswith('vts'):
        reader = vtk.vtkXMLStructuredGridReader()
    elif fileName.endswith('vtu'):
        reader = vtk.vtkXMLUnstructuredGridReader()
    elif fileName.endswith('stl'):
        reader = vtk.vtkSTLReader()
        reader.ScalarTagsOn()
    elif fileName.endswith('nii') or fileName.endswith('nii.gz'):
        reader = vtk.vtkNIFTIImageReader()
    elif fileName.endswith('vti'):
        reader = vtk.vtkXMLImageDataReader()
    elif fileName.endswith('vtk'):
        reader = vtk.vtkPolyDataReader()
    elif fileName.endswith('vtm'):
        reader = vtk.vtkXMLMultiBlockDataReader()
    elif fileName.endswith('nrrd'):
        reader = vtk.vtkNrrdReader()
    elif fileName.endswith('mha') or fileName.endswith('mhd'):
        reader = vtk.vtkMetaImageReader()
    elif fileName.endswith('png'):
        reader = vtk.vtkPNGReader()
    elif fileName.endswith('jpg') or fileName.endswith('jpeg'):
        reader = vtk.vtkJPEGReader()
    elif fileName.endswith('tif') or fileName.endswith('tiff'):
        reader = vtk.vtkTIFFReader()
    elif fileName.endswith('ply'):
        reader = vtk.vtkPLYReader()
    elif fileName.endswith('pvd'):
        raise IOError(' PVD - should use readPVD()')
    else:
        raise IOError(fileName + ' not correct extension')
    reader.SetFileName(fileName)
    reader.Update()
    return reader.GetOutput()


# =========================================================================
def writeNifti(data: vtk.vtkDataObject, fileName: str) -> str:
    """
    Write a NIfTI file.

    Args:
        data (vtk.vtkDataObject): The VTK data object to write.
        fileName (str): The path of the NIfTI file to write to.

    Returns:
        str: The path of the NIfTI file written to.
    """
    writer = vtk.vtkNIFTIImageWriter()
    writer.SetFileName(fileName)
    writer.SetInputData(data)
    writer.Write()
    return fileName

def readNifti(fileName: str) -> vtk.vtkDataObject:
    """
    Read a NIfTI file.

    Args:
        fileName (str): The path of the NIfTI file to read.

    Returns:
        vtk.vtkDataObject: The VTK data object.
    """
    reader = vtk.vtkNIFTIImageReader()
    reader.SetFileName(fileName)
    reader.Update()
    return reader.GetOutput()

def readNRRD(fileName: str) -> vtk.vtkDataObject:
    """
    Read a NRRD file.

    Args:
        fileName (str): The path of the NRRD file to read.

    Returns:
        vtk.vtkDataObject: The VTK data object.
    """
    reader = vtk.vtkNrrdReader()
    reader.SetFileName(fileName)
    reader.Update()
    return reader.GetOutput()
# =========================================================================


# =========================================================================
##          PVD Stuff
# =========================================================================
def _writePVD(rootDirectory: str, filePrefix: str, outputSummary: Dict[int, Dict[str, Union[str, float]]]) -> str:
    """
    Write a PVD file. For internal use only.

    Args:
        rootDirectory (str): The root directory.
        filePrefix (str): The file prefix.
        outputSummary (Dict[int, Dict[str, Union[str, float]]]): The output summary.
    """
    fileOut = os.path.join(rootDirectory, filePrefix + '.pvd')
    with open(fileOut, 'w') as f:
        f.write('<?xml version="1.0"?>\n')
        f.write('<VTKFile type="Collection" version="0.1" byte_order="LittleEndian">\n')
        f.write('<Collection>\n')
        for timeId in sorted(outputSummary.keys()):
            sTrueTime = outputSummary[timeId]['TrueTime']
            tFileName = str(outputSummary[timeId]['FileName'])
            f.write('<DataSet timestep="%7.5f" file="%s"/>\n' % (sTrueTime, tFileName))
        f.write('</Collection>\n')
        f.write('</VTKFile>')
    return fileOut


def _makePVDOutputDict(vtkDict: Dict[Union[str, float], vtk.vtkDataObject], filePrefix: str, fileExtn: str, subDir: str = '') -> Dict[int, Dict[str, Union[str, float]]]:
    """
    Make a PVD output dictionary. For internal use only.

    Args:
        vtkDict (Dict[Union[str, float], vtk.vtkDataObject]): The dictionary of VTK data objects.
        filePrefix (str): The file prefix.
        fileExtn (str): The file extension.
        subDir (str, optional): The subdirectory. Defaults to ''.

    Returns:
        Dict[int, Dict[str, Union[str, float]]]: The output summary.
    """
    outputSummary = {}
    myKeys = vtkDict.keys()
    myKeys = sorted(myKeys)
    for timeId in range(len(myKeys)):
        fileName = __buildFileName(filePrefix, timeId, fileExtn)
        trueTime = myKeys[timeId]
        outputMeta = {'FileName': os.path.join(subDir, fileName), 'TimeID': timeId, 'TrueTime': trueTime}
        outputSummary[timeId] = outputMeta
    return outputSummary

def __writePVDData(vtkDict: Dict[Union[str, float], vtk.vtkDataObject], rootDir: str, filePrefix: str, fileExtn: str, subDir: str = '') -> None:
    """
    Write PVD data. For internal use only.

    Args:
        vtkDict (Dict[Union[str, float], vtk.vtkDataObject]): The dictionary of VTK data objects.
        rootDir (str): The root directory.
        filePrefix (str): The file prefix.
        fileExtn (str): The file extension.
        subDir (str, optional): The subdirectory. Defaults to ''.
    """
    myKeys = vtkDict.keys()
    myKeys = sorted(myKeys)
    for timeId in range(len(myKeys)):
        fileName = __buildFileName(filePrefix, timeId, fileExtn)
        fileOut = os.path.join(rootDir, subDir, fileName)
        if type(vtkDict[myKeys[timeId]]) == str:
            os.rename(vtkDict[myKeys[timeId]], fileOut)
        else:
            writeVTKFile(vtkDict[myKeys[timeId]], fileOut)

def writeVTK_PVD_Dict(vtkDict: Dict[Union[str, float], vtk.vtkDataObject], rootDir: str, filePrefix: str, fileExtn: str, BUILD_SUBDIR: bool = True) -> str:
    """
    Write dict of time:vtkObj to pvd file
        If dict is time:fileName then will copy files

    Args:
        vtkDict (Dict[Union[str, float], vtk.vtkDataObject]): The dictionary of VTK data objects.
        rootDir (str): The root directory.
        filePrefix (str): make filePrefix.pvd
        fileExtn (str): file extension (e.g. vtp, vti, vts etc)
        BUILD_SUBDIR (bool, optional): to build subdir (filePrefix.pvd in root, then data in root/filePrefix/). Defaults to True.
    
    Returns:
        str: full file name
    """
    filePrefix = os.path.splitext(filePrefix)[0]
    subDir = ''
    fullPVD = os.path.join(rootDir, checkIfExtnPresent(filePrefix, 'pvd'))
    if os.path.isfile(fullPVD) & (type(list(vtkDict.values())[0]) != str):
        deleteFilesByPVD(fullPVD, QUIET=True)
    if BUILD_SUBDIR:
        subDir = filePrefix
        if not os.path.isdir(os.path.join(rootDir, subDir)):
            os.mkdir(os.path.join(rootDir, subDir))
    outputSummary = _makePVDOutputDict(vtkDict, filePrefix, fileExtn, subDir)
    __writePVDData(vtkDict, rootDir, filePrefix, fileExtn, subDir)
    return _writePVD(rootDir, filePrefix, outputSummary)


def pvdAddTimeToFieldData(pvdf_or_vtkDict: Union[str, Dict[Union[str, float], vtk.vtkDataObject]]) -> Union[str, Dict[Union[str, float], vtk.vtkDataObject]]:
    """
    Add time to field data.

    Args:
        pvdf_or_vtkDict (Union[str, Dict[Union[str, float], vtk.vtkDataObject]]): The PVD file path or dictionary of VTK data objects.

    Returns:
        Union[str, Dict[Union[str, float], vtk.vtkDataObject]]: The PVD file path or dictionary of VTK data objects.
    """
    rr, ff, ee = pvdGetDataFileRoot_Prefix_and_Ext(pvdf_or_vtkDict)
    pvdf_or_vtkDict = readPVD(pvdf_or_vtkDict)
    for iTime in sorted(pvdf_or_vtkDict.keys()):
        timeArray = vtk.vtkFloatArray()
        timeArray.SetNumberOfValues(1)
        timeArray.SetValue(0, iTime)
        timeArray.SetName("Time")
        pvdf_or_vtkDict[iTime].GetFieldData().AddArray(timeArray)
    if rr is not None:
        return writeVTK_PVD_Dict(pvdf_or_vtkDict, rr, ff, ee)
    else:
        return pvdf_or_vtkDict



def writeZipFromVTK_PVD_Dict(vtkDict: Dict[Union[str, float], vtk.vtkDataObject], fileExt: str, zipFileOut: str) -> None:
    """
    Write a ZIP file from a dictionary of VTK data objects.

    Args:
        vtkDict (Dict[Union[str, float], vtk.vtkDataObject]): The dictionary of VTK data objects.
        fileExt (str): The file extension.
        zipFileOut (str): The output ZIP file path.
    """
    # first save times to field data
    pvdAddTimeToFieldData(vtkDict)
    # write as normal, then zip, then del.
    r, f = os.path.split(zipFileOut)
    f = os.path.splitext(f)[0]
    pvdOut = writeVTK_PVD_Dict(vtkDict, r, f, fileExt)
    if zipFileOut[-4:] == '.zip':
        zipFileOut = zipFileOut[:-4]
    shutil.make_archive(zipFileOut, 'zip', os.path.join(r, f))
    deleteFilesByPVD(pvdOut)


def deleteFilesByPVD(pvdFile: str, FILE_ONLY: bool = False, QUIET: bool = False) -> int:
    """
    Delete files referenced by a PVD file.

    Args:
        pvdFile (str): The PVD file path.
        FILE_ONLY (bool, optional): Whether to delete only the files. Defaults to False.
        QUIET (bool, optional): Whether to suppress warnings. Defaults to False.

    Returns:
        int: 0 if successful, 1 if error.
    """
    if FILE_ONLY:
        try:
            os.remove(pvdFile)
        except (IOError, OSError):
            print('    warning - file not found %s' % (pvdFile))
            return 1
        return 0
    try:
        pvdDict = readPVDFileName(pvdFile)
        for iKey in pvdDict.keys():
            try:
                os.remove(pvdDict[iKey])
            except OSError:
                pass  # ignore this as may be shared by and deleted by another pvd
        os.remove(pvdFile)
    except (IOError, OSError):
        if (not QUIET)&("pvd" not in pvdFile):
            print('    warning - file not found %s' % (pvdFile))
    try:
        head, _ = os.path.splitext(pvdFile)
        os.rmdir(head)
    except (IOError, OSError):
        if not QUIET:
            print('    warning - dir not found %s' % (head))
    return 0


def readPVDFileName(fileIn: str, vtpTime: float = 0.0, timeIDs: List[int] = [], RETURN_OBJECTS_DICT: bool = False) -> Union[Dict[float, str], Dict[float, vtk.vtkDataObject]]:
    """
    Read PVD file, return dictionary of fullFileNames - keys = time
    So DOES NOT read file
    If not pvd - will return dict of {0.0 : fileName}
    
    Args:
        fileIn: str - PVD file path
        vtpTime: float - time to read
        timeIDs: list of int - time IDs to read
        RETURN_OBJECTS_DICT: bool - return dictionary of vtkDataObjects
    
    Returns:
        dict of fullFileNames - keys = time
    """
    _, ext = os.path.splitext(fileIn)
    if ext != '.pvd':
        if RETURN_OBJECTS_DICT:
            return {vtpTime: readVTKFile(fileIn)}
        else:
            return {vtpTime: fileIn}
    #
    vtkDict = pvdGetDict(fileIn, timeIDs)
    if RETURN_OBJECTS_DICT:
        kk = vtkDict.keys()
        return dict(zip(kk, [readVTKFile(vtkDict[i]) for i in kk]))
    else:
        return vtkDict

def readImageFileToDict(imageFile: str) -> Dict[float, vtk.vtkDataObject]:
    """
    Read an image file and return a dictionary of VTK data objects.

    Args:
        imageFile (str): The image file path.

    Returns:
        Dict[float, vtk.vtkDataObject]: A dictionary of VTK data objects (keys = time).
    """
    return readPVD(imageFile)

def readPVD(fileIn: str, timeIDs: List[int] = []) -> Dict[float, vtk.vtkDataObject]:
    """
    Read a PVD file and return a dictionary of VTK data objects.

    Args:
        fileIn (str): The PVD file path.
        timeIDs (List[int], optional): A list of time step IDs to include. Defaults to an empty list.

    Returns:
        Dict[float, vtk.vtkDataObject]: A dictionary of VTK data objects (keys = time).
    """
    return readPVDFileName(fileIn, timeIDs=timeIDs, RETURN_OBJECTS_DICT=True)

def pvdGetDict(pvd: Union[str, ET.Element, Dict[float, str]], timeIDs: List[int] = []) -> Dict[float, str]:
    """
    Get a dictionary of time steps and their corresponding file names from a PVD file.

    Args:
        pvd (Union[str, ET.Element, Dict[float, str]]): The PVD file path, XML element, or dictionary.
        timeIDs (List[int], optional): A list of time step IDs to include. Defaults to an empty list.

    Returns:
        Dict[float, str]: A dictionary mapping time steps to file names.
    """
    if type(pvd) == str:
        root = ET.parse(pvd).getroot()
    elif type(pvd) == dict:
        return pvd
    else:
        root = pvd
    nTSteps = len(root[0])
    if len(timeIDs) == 0:
        timeIDs = range(nTSteps)
    else:
        for k1 in range(len(timeIDs)):
            if timeIDs[k1] < 0:
                timeIDs[k1] = nTSteps + timeIDs[k1]
    pvdTimesFilesDict = {}
    rootDir = os.path.dirname(pvd)
    for k in range(nTSteps):
        if k not in timeIDs:
            continue
        a = root[0][k].attrib
        fullVtkFileName = os.path.join(rootDir, a['file'])
        pvdTimesFilesDict[float(a['timestep'])] = fullVtkFileName
    return pvdTimesFilesDict

def pvdGetClosestTimeToT(pvd: Union[str, ET.Element, Dict[float, str]], T: float) -> float:
    """
    Get the time step closest to a given time from a PVD file.

    Args:
        pvd (Union[str, ET.Element, Dict[float, str]]): The PVD file path, XML element, or dictionary.
        T (float): The target time.

    Returns:
        float: The closest time step.
    """
    pvdDict = pvdGetDict(pvd)
    tt = list(sorted((pvdDict.keys())))
    return tt[pvdGetIdOfT(pvdDict, T)]

def pvdGetTimes(pvd: Union[str, ET.Element, Dict[float, str]]) -> List[float]:
    """
    Get a list of time steps from a PVD file.

    Args:
        pvd (Union[str, ET.Element, Dict[float, str]]): The PVD file path, XML element, or dictionary.

    Returns:
        List[float]: A list of time steps.
    """
    pvdDict = pvdGetDict(pvd)
    tt = list(sorted((pvdDict.keys())))
    return tt

def pvdGetIdOfT(pvd: Union[str, ET.Element, Dict[float, str]], T: float) -> int:
    """
    Get the ID of the time step closest to a given time from a PVD file.

    Args:
        pvd (Union[str, ET.Element, Dict[float, str]]): The PVD file path, XML element, or dictionary.
        T (float): The target time.

    Returns:
        int: The ID of the closest time step.
    """
    pvdDict = pvdGetDict(pvd)
    return np.argmin(abs(np.array(sorted(pvdDict.keys()))-T))

def pvdGetFileAtT(pvd: Union[str, ET.Element, Dict[float, str]], T: float) -> str:
    """
    Get the file name corresponding to the time step closest to a given time from a PVD file.

    Args:
        pvd (Union[str, ET.Element, Dict[float, str]]): The PVD file path, XML element, or dictionary.
        T (float): The target time.

    Returns:
        str: The file name corresponding to the closest time step.
    """
    pvdDict = pvdGetDict(pvd)
    return pvdDict[pvdGetClosestTimeToT(pvdDict, T)]

def pvdGetNumberTimePoints(pvd: Union[str, ET.Element, Dict[float, str]]) -> int:
    """
    Get the number of time points in a PVD file.

    Args:
        pvd (Union[str, ET.Element, Dict[float, str]]): The PVD file path, XML element, or dictionary.

    Returns:
        int: The number of time points.
    """
    return len(readPVDFileName(pvd).keys())

def pvdGetDataFileRoot_Prefix_and_Ext(pvd: Union[str, ET.Element, Dict[float, str]]) -> Tuple[str, str, str]:
    """
    Get the root directory, prefix, and extension of the data files referenced in a PVD file.

    Args:
        pvd (Union[str, ET.Element, Dict[float, str]]): The PVD file path, XML element, or dictionary.

    Returns:
        Tuple[str, str, str]: A tuple containing the root directory, prefix, and extension.
    """
    h, t = os.path.split(pvd)
    t = os.path.splitext(t)[0]
    try:
        ff = pvdGetFileAtT(pvd, 0)
    except FileNotFoundError:
        ff = pvd
    return h, t, os.path.splitext(ff)[1]

def pvdGetDataClosestTo(pvd: Union[str, ET.Element, Dict[float, str]], refT: float) -> vtk.vtkDataObject:
    """
    Get the data object corresponding to the time step closest to a given reference time from a PVD file.

    Args:
        pvd (Union[str, ET.Element, Dict[float, str]]): The PVD file path, XML element, or dictionary.
        refT (float): The reference time.

    Returns:
        vtk.vtkDataObject: The data object corresponding to the closest time step.
    """
    closestT = pvdGetClosestTimeToT(pvd, refT)
    xID = pvdGetIdOfT(pvd, closestT)
    dd = readPVD(pvd, [xID])
    return dd[closestT]

def pvdMultiply(pvdFile: str, nCycles: int, outputFileName: Optional[str] = None) -> str:
    """
    Multiply the time steps in a PVD file by a given number of cycles.

    Args:
        pvdFile (str): The input PVD file path.
        nCycles (int): The number of cycles to multiply the time steps by.
        outputFileName (Optional[str], optional): The output PVD file path. Defaults to None.

    Returns:
        str: The output PVD file path.
    """
    pvdTree = ET.parse(pvdFile)
    collectionTag = pvdTree.find('Collection')
    fileList = []
    timeList = []
    for elem in pvdTree.iter(tag='DataSet'):
        timeList.append(float(elem.attrib['timestep']))
        fileList.append(elem.attrib['file'])
    fileList = [x for (y, x) in sorted(zip(timeList, fileList))]  # @UnusedVariable
    timeList.sort()
    nTS = len(timeList)
    nominalInterval = float(timeList[-1]) + (float(timeList[-1]) - float(timeList[-2]))
    timeMax = (nCycles + 1) * nominalInterval
    try:
        if not (nCycles).is_integer():
            timeMax = nCycles * nominalInterval
            nCycles = int(np.ceil(nCycles))
        else:
            nCycles = int(nCycles)
    except AttributeError:
        pass
    for k1 in range(1, nCycles):
        for k2 in range(nTS):
            newTime = (nominalInterval * k1) + float(timeList[k2])
            timeList.append(str(newTime))
            fileList.append(fileList[k2])
    for i in range(nTS, nTS * (nCycles)):
        elem = ET.Element("DataSet")
        if float(timeList[i]) < timeMax:
            elem.attrib["file"] = fileList[i]
            elem.attrib["timestep"] = timeList[i]
            collectionTag.append(elem)
    if outputFileName is None:
        outputFileName = pvdFile[:-4] + '_' + str(nCycles) + 'cycles.pvd'
    collectionTag = pvdTree.find('Collection');xmlIndent(collectionTag)
    pvdTree.write(outputFileName, 'UTF-8', xml_declaration=True)
    return outputFileName

def pvdReverse(pvdFileIn: str, suffix: str = "REV", outfile: Optional[str] = None) -> str:
    """
    Reverse the time steps in a PVD file.

    Args:
        pvdFileIn (str): The input PVD file path.
        suffix (str, optional): The suffix to append to the output file name. Defaults to "REV".
        outfile (Optional[str], optional): The output file path. Defaults to None.

    Returns:
        str: The output PVD file path.
    """
    pvdFileOut = pvdFileIn[:-4] + suffix + '.pvd'
    pvdTree = ET.parse(pvdFileIn)
    root = pvdTree.find('Collection')
    fileList = []
    timeList = []
    for i in root:
        fileList.append(i.attrib['file'])
        timeList.append(i.attrib['timestep'])
    sortId = np.argsort(np.array(timeList))
    sortRevId = sortId[::-1]
    for i in range(len(root)):
        root[i].attrib['timestep'] = timeList[i]
        root[i].attrib['file'] = fileList[sortRevId[i]]
    root.set('updated', 'yes')
    collectionTag = pvdTree.find('Collection');
    xmlIndent(collectionTag)
    pvdTree.write(pvdFileOut, 'UTF-8', xml_declaration=True)
    if outfile is not None:
        pvdD = readPVD(pvdFileOut)
        _,_,e = pvdGetDataFileRoot_Prefix_and_Ext(pvdFileOut)
        outR, outP = os.path.split(outfile)
        outP = os.path.splitext(outP)[0]
        deleteFilesByPVD(pvdFileOut,FILE_ONLY=True)
        pvdFileOut = writeVTK_PVD_Dict(pvdD, outR, outP, e, True)
    return pvdFileOut

def pvdRestart_t(pvdFile: str, timeStart: float, outputFile: Optional[str] = None) -> str:
    """
    Restart a PVD file from a given time.

    Args:
        pvdFile (str): The input PVD file path.
        timeStart (float): The start time.
        outputFile (Optional[str], optional): The output PVD file path. Defaults to None.

    Returns:
        str: The output PVD file path.
    """
    idStart = pvdGetIdOfT(pvdFile, timeStart)
    return pvdRestart(pvdFile, idStart, outputFile)

def pvdRestart(pvdFile: str, idStart: int, outputFile: Optional[str] = None) -> str:
    """
    Restart a PVD file from a given time step ID.

    Args:
        pvdFile (str): The input PVD file path.
        idStart (int): The start time step ID.
        outputFile (Optional[str], optional): The output PVD file path. Defaults to None.

    Returns:
        str: The output PVD file path.
    """
    pvdTree = ET.parse(pvdFile)
    collectionTag = pvdTree.find('Collection')
    count = -1
    for iDataset in collectionTag.findall('DataSet'):
        count += 1
        if count < idStart:
            collectionTag.remove(iDataset)
    if not outputFile:
        outputFile = pvdFile[:-4] + '_' + str(idStart) + 'start.pvd'
    collectionTag = pvdTree.find('Collection')
    xmlIndent(collectionTag)
    pvdTree.write(outputFile, 'UTF-8', xml_declaration=True)
    return outputFile

def pvdClip(pvdFile: str, idEnd: int, outputFile: Optional[str] = None) -> str:
    """
    Clip a PVD file to a given time step ID.

    Args:
        pvdFile (str): The input PVD file path.
        idEnd (int): The end time step ID.
        outputFile (Optional[str], optional): The output PVD file path. Defaults to None.

    Returns:
        str: The output PVD file path.
    """
    pvdTree = ET.parse(pvdFile)
    collectionTag = pvdTree.find('Collection')
    count = -1
    for iDataset in collectionTag.findall('DataSet'):
        count += 1
        if count > idEnd:
            collectionTag.remove(iDataset)
    if not outputFile:
        outputFile = pvdFile[:-4] + '_' + str(idEnd) + 'end.pvd'
    collectionTag = pvdTree.find('Collection')
    xmlIndent(collectionTag)
    pvdTree.write(outputFile, 'UTF-8', xml_declaration=True)
    return outputFile

def pvdResetTimes(pvdFile: str, startTime: float, endTime: float, pvdFileOut: str) -> str:
    """
    Reset the time steps in a PVD file to a given range.

    Args:
        pvdFile (str): The input PVD file path.
        startTime (float): The start time.
        endTime (float): The end time.
        pvdFileOut (str): The output PVD file path.

    Returns:
        str: The output PVD file path.
    """
    pvdTree = ET.parse(pvdFile)
    root = pvdTree.find('Collection')
    fileList = []
    timeList = []
    newTimes = np.linspace(startTime, endTime, len(root))
    for k1, i in enumerate(root):
        fileList.append(i.attrib['file'])
        timeList.append('%5.5f'%(newTimes[k1]))
    for i in range(len(root)):
        root[i].attrib['timestep'] = timeList[i]
        root[i].attrib['file'] = fileList[i]
    root.set('updated', 'yes')
    collectionTag = pvdTree.find('Collection');
    xmlIndent(collectionTag)
    pvdTree.write(pvdFileOut, 'UTF-8', xml_declaration=True)
    return pvdFileOut

def pvdResetStartPoint_Id(pvdFile: str, startID: int, pvdFileOut: str) -> str:
    """
    Reset the start point of a PVD file to a given time step ID.

    Args:
        pvdFile (str): The input PVD file path.
        startID (int): The start time step ID.
        pvdFileOut (str): The output PVD file path.

    Returns:
        str: The output PVD file path.
    """
    times = list(readPVDFileName(pvdFile).keys())
    startTime =times[startID]
    return pvdResetStartPoint(pvdFile, startTime, pvdFileOut, QUIET=True)

def pvdResetStartPoint(pvdFile: str, startTime: float, pvdFileOut: str, QUIET: bool = True) -> str:
    """
    Reset the start point of a PVD file to a given time.

    Args:
        pvdFile (str): The input PVD file path.
        startTime (float): The start time.
        pvdFileOut (str): The output PVD file path.
        QUIET (bool, optional): Whether to suppress output. Defaults to True.

    Returns:
        str: The output PVD file path.
    """
    pvdTree = ET.parse(pvdFile)
    root = pvdTree.find('Collection')
    fileList = []
    timeList = []
    for i in root:
        fileList.append(i.attrib['file'])
        timeList.append(float(i.attrib['timestep']))
    idStart = int(np.argmin([abs(i-startTime) for i in timeList]))
    if not QUIET:
        print('Ressetting start ID to %d'%(idStart))
    for k1, ID in enumerate(range(idStart, len(root))):
        root[k1].attrib['timestep'] = '%5.5f'%(timeList[k1])
        root[k1].attrib['file'] = fileList[ID]
    for k2 in range(idStart):
        root[k1+1+k2].attrib['timestep'] = '%5.5f'%(timeList[k2+k1+1])
        root[k1+1+k2].attrib['file'] = fileList[k2]
    root.set('updated', 'yes')
    collectionTag = pvdTree.find('Collection');
    xmlIndent(collectionTag)
    pvdTree.write(pvdFileOut, 'UTF-8', xml_declaration=True)
    return pvdFileOut

def pvdAddStartPointAtEndForPeriodicy(pvdFile: str, pvdFileOut: str) -> str:
    """
    Add a start point at the end of a PVD file for periodicity.

    Args:
        pvdFile (str): The input PVD file path.
        pvdFileOut (str): The output PVD file path.

    Returns:
        str: The output PVD file path.
    """
    pvdTree = ET.parse(pvdFile)
    root = pvdTree.find('Collection')
    nSteps = len(root)
    fileList = []
    timeList = []
    for i in root:
        fileList.append(i.attrib['file'])
        timeList.append(float(i.attrib['timestep']))
    fileList.append(root[0].attrib['file'])
    lastTime = timeList[-1] + timeList[-1] - timeList[-2]
    timeList.append(lastTime)
    #
    for iDataset in root.findall('DataSet'):
        root.remove(iDataset)
    #
    for k1 in range(len(fileList)):
        elem = ET.Element("DataSet")
        elem.attrib["file"] = fileList[k1]
        elem.attrib["timestep"] = '%5.5f'%(timeList[k1])
        root.append(elem)
    root.set('updated', 'yes')
    collectionTag = pvdTree.find('Collection');xmlIndent(collectionTag)
    pvdTree.write(pvdFileOut, 'UTF-8', xml_declaration=True)
    return pvdFileOut

def buildPVD_WaterFallRestarts(pvdFileIn: str, pvdFileOut: str, restartEvery: int, nCycle: int = 1, TO_REV: bool = False) -> List[str]:
    """
    Build a waterfall of restarted PVD files.

    Args:
        pvdFileIn (str): The input PVD file path.
        pvdFileOut (str): The output PVD file path.
        restartEvery (int): The number of time steps between restarts.
        nCycle (int, optional): The number of cycles. Defaults to 1.
        TO_REV (bool, optional): Whether to reverse the time steps. Defaults to False.

    Returns:
        List[str]: A list of output PVD file paths.
    """
    h, pvdFileOut = os.path.split(pvdFileOut)
    pvdFileOut = os.path.splitext(pvdFileOut)[0]
    pvdNCycleFile = pvdMultiply(pvdFileIn, nCycle)
    restartIds = range(0, len(readPVDFileName(pvdNCycleFile))-1, restartEvery)
    allFilesList = []
    outputSummary = {}
    for k1 in restartIds:
        if TO_REV:
            ttf = pvdNCycleFile[:-4]+'-R.pvd'
            ttf = pvdRestart(ttf, int(k1))
            outputFile = pvdReverse(ttf, outfile=ttf)
        else:
            outputFile = pvdRestart(pvdNCycleFile, int(k1))
        outputSummary[int(k1)] = {'TrueTime':int(k1), 'FileName':os.path.split(outputFile)[1]}
        allFilesList.append(outputFile)
    wrappingPVDFile = _writePVD(h, pvdFileOut, outputSummary)
    return allFilesList




"""
@author Fraser Callaghan
"""

import numpy as np
import itertools
from scipy import stats
from scipy import optimize, interpolate
from scipy.spatial import cKDTree
import vtk


# ======================================================================================================================
#           MATH
# ======================================================================================================================
def getIDOfClosestFloat(iFloat, floatList):
    return np.argmin([abs(i-iFloat) for i in floatList])


def getClosestFloat(iFloat, floatList):
    return floatList[getIDOfClosestFloat(iFloat, floatList)]


def getIDOfClosestPoint(point, points):
    dist_2 = squareDistPointPoints(point, points)
    return np.argmin(dist_2)


def getClosestPoint(point, points):
    return points[getIDOfClosestPoint(point, points)]


def distPointPoints(point, points):
    dist_2 = squareDistPointPoints(point, points)
    return np.sqrt(dist_2)


def squareDistPointPoints(point, points):
    points = np.asarray(points)
    if points.ndim == 1:
        points = points[np.newaxis, :]
    deltas = points - point
    dist_2 = np.einsum('ij,ij->i', deltas, deltas)
    return dist_2


def sortedIDsOfClosestPoints(point, points):
    dist_2 = squareDistPointPoints(point, points)
    return np.argsort(dist_2)


def distTwoPoints(a, b):
    return np.sqrt(np.sum((np.asarray(a) - np.asarray(b)) ** 2))


def getIDsOfClosestPoints(pointsA, pointsB):
    pointsA = np.asarray(pointsA)
    pointsB = np.asarray(pointsB)
    distances = np.sum((pointsA[:, np.newaxis, :] - pointsB[np.newaxis, :, :]) ** 2, axis=2)
    minIDa, minIDb = np.unravel_index(np.argmin(distances), distances.shape)
    return minIDa, minIDb


def getIDsOfFartherestPoints(pointsA, pointsB):
    pointsA = np.asarray(pointsA)
    pointsB = np.asarray(pointsB)
    distances = np.sum((pointsA[:, np.newaxis, :] - pointsB[np.newaxis, :, :]) ** 2, axis=2)
    maxIDa, maxIDb = np.unravel_index(np.argmax(distances), distances.shape)
    return maxIDa, maxIDb


def vectorMagnitudes(vecs):
    vecs = np.asarray(vecs)
    try:
        return np.sqrt(np.sum(vecs*vecs,-1))
    except IndexError:
        return np.array([np.linalg.norm(vecs)])


def normaliseArray(vectors):
    """
    Normalise an array of vectors.

    Parameters:
    vectors (array-like): Input array of vectors. Can be 1D or 2D.

    Returns:
    numpy.ndarray: Array of normalised vectors.
    """
    vectors = np.asarray(vectors)
    if vectors.ndim == 1:
        vectors = vectors[np.newaxis, :]
    magnitudes = np.linalg.norm(vectors, axis=-1, keepdims=True)
    magnitudes = np.maximum(magnitudes, 1e-8) # Avoid division by zero
    normalised = vectors / magnitudes
    
    return normalised.squeeze()


def angleBetween2Vec(v1, v2, RETURN_DEGREES=False):
    """
    Returns the angle in radians between vectors 'v1' and 'v2'::
        Can be lists of same size
    :param v1:
    :param v2:
    :param RETURN_DEGREES:
    :return: angle in radians
    """
    if RETURN_DEGREES:
        return vtk.vtkMath.DegreesFromRadians(angleBetween2Vec(v1, v2, RETURN_DEGREES=False))
    return vtk.vtkMath.AngleBetweenVectors(v1, v2)


def angleBetweenVecAndPlane(vector, planeNormal):
    return np.arcsin(abs(np.dot(vector, planeNormal)) / (np.linalg.norm(vector)) * np.linalg.norm(planeNormal))


def areVecsMatching(v1, v2, angleTol_rad, magTol):
    m1, m2 = np.linalg.norm(v1), np.linalg.norm(v2)
    if abs(m1 - m2) > magTol:
        return False
    theta = angleBetween2Vec(v1, v2)
    if theta > angleTol_rad:
        return False
    return True


def __forcePts_nx3(pts):
    pts = np.array(pts)
    dims = pts.shape
    if len(dims) == 1:
        pts = np.reshape(pts, (1,dims[0]))
    else:
        if ((dims[0] == 3) and (dims[1] != 3)) | ((dims[0] == 2) and (dims[1] > 3)):
            pts = pts.T
    return pts


# ======================================================================================================================
#           FITTING & PROJECTION
# ======================================================================================================================
def linearFit(X ,Y):
    """
    Fit a linear model to the data.

    Parameters:
    - X: Independent variable.
    - Y: Dependent variable.

    Returns:
    slope, intercept, R, pValue, std: Parameters of the linear model.
    """
    slope, intercept, R, pValue, std = stats.linregress(X, Y) # c, m, R
    return slope, intercept, R, pValue, std


def fitPlaneToPoints(pts):
    """
    Fit a plane to a set of 3D points.

    Parameters:
    - pts: A numpy array of shape (N, 3) representing the 3D points.

    Returns:
    - planeABC: Coefficients of the plane equation [a, b, c, d].
    """
    pts = __forcePts_nx3(pts)
    A = pts - np.mean(pts, 0)
    _, S, V = np.linalg.svd(A)
    i = np.argmin(S)  # find position of minimal singular value
    coeff = V[i, :]  # this may be multiple
    planeABC = coeff / np.linalg.norm(coeff[:3])
    Ds = [planeABC[0]*x[0]+planeABC[1]*x[1]+planeABC[2]*x[2] for x in pts]
    return np.hstack((planeABC, -np.mean(Ds)))


def polyfit2d(x, y, z, order=3):
    ncols = (order + 1) ** 2
    G = np.zeros((x.size, ncols))
    ij = itertools.product(range(order + 1), range(order + 1))
    for k, (i, j) in enumerate(ij):
        G[:, k] = x ** i * y ** j
    m, _, _, _ = np.linalg.lstsq(G, z, rcond=None)
    return m


def polyval2d(x, y, m):
    order = int(np.sqrt(len(m))) - 1
    ij = itertools.product(range(order + 1), range(order + 1))
    z = np.zeros_like(x)
    for a, (i, j) in zip(m, ij):
        z += a * x ** i * y ** j
    return z


def projectPtsToPlane(pts, plane):
    """
    projects the points so all lie on the plane
    :param pts: list/array of points Nx3
    :param plane: equation of plane: either [a,b,c,d] or [norm, ptInPlane]
    :return: pts array all lying on plane
    """
    if isinstance(plane, tuple):
        planeNorm = plane[0]
        inplanePt = plane[1]
        D = -inplanePt[0] * planeNorm[0] - inplanePt[1] * planeNorm[1] - \
            inplanePt[2] * planeNorm[2]
    else:
        planeNorm = plane[:3]
        D = plane[3]
    pts = __forcePts_nx3(pts)
    fraction = (planeNorm[0] * pts[:, 0] +
                planeNorm[1] * pts[:, 1] +
                planeNorm[2] * pts[:, 2] + D) / \
               (planeNorm[0] ** 2 + planeNorm[1] ** 2 + planeNorm[2] ** 2)
    x0 = pts[:, 0] - planeNorm[0] * fraction
    y0 = pts[:, 1] - planeNorm[1] * fraction
    z0 = pts[:, 2] - planeNorm[2] * fraction
    return np.column_stack((x0, y0, z0))


def project3DPointsToPlanarCoordinateSystem(pts, planeNorm=None):
    origin = np.mean(pts, axis=0)
    normal = planeNorm / np.linalg.norm(planeNorm)
    projected_points = projectPtsToPlane(pts, (planeNorm, origin))
    # Create a local coordinate system on the plane
    # Find two vectors in the plane
    u = projected_points[1] - projected_points[0]
    u = u / np.linalg.norm(u)
    v = np.cross(normal, u)
    # Convert the projected 3D points to 2D coordinates in the (u, v) plane
    points_2d = np.array([[np.dot(pt - origin, u), np.dot(pt - origin, v)] for pt in projected_points])
    return points_2d, u, v, origin


def project3DPointsToPlanarCoordinateSystem_OLD(pts, planeNorm=None):
    """
    Will rotate pts to lie in given plane
       - shifts pts to be centered on origin
        FIXME: If points are aligned with an axis (almost) then will fail.
       - raise LinAlgError
          Need to check for this, then simply rotate by different vector set.
    source: http://mathforum.org/library/drmath/view/51727.html
    source: https://stackoverflow.com/questions/23472048/projecting-3d-points-to-2d-plane
    :param pts:
    :param planeNorm:
    :return:
    """
    pts = __forcePts_nx3(pts)
    npts, _ = pts.shape
    try:
        len(planeNorm)
    except TypeError:
        planeNorm = fitPlaneToPoints(pts)[:3]
    planeNorm = planeNorm / np.linalg.norm(planeNorm)
    cPt = np.mean(pts, 0)
    pPts = projectPtsToPlane(pts, (planeNorm, cPt))
    e1 = np.array([-1 * planeNorm[1], planeNorm[0], 0.0])
    e1 = e1 / np.linalg.norm(e1)
    e2 = np.cross(planeNorm, e1)
    A = np.array([e1[:2], e2[:2]])
    xPrime = np.zeros((npts, 2))
    for k1 in range(npts):
        b = pPts[k1] - cPt
        xPrime[k1, :] = np.linalg.solve(A.T, b[:2])
    return xPrime, e1, e2, cPt


def project3DPointsToPlanarCoordinateSystem2(pts, planeNorm=None, datumPt=None):
    # https://stackoverflow.com/questions/26369618/getting-local-2d-coordinates-of-vertices-of-a-planar-polygon-in-3d-space
    pts = __forcePts_nx3(pts)
    npts, _ = pts.shape
    if planeNorm is None:
        planeNorm = fitPlaneToPoints(pts)[:3]
    if datumPt is None:
        datumPt = np.mean(pts,0)
    locx = pts[0] - datumPt
    locy = np.cross(planeNorm, locx)
    locx = locx / np.linalg.norm(locx)
    locy = locy / np.linalg.norm(locy)
    xy = [[np.dot(p-datumPt,locx), np.dot(p-datumPt,locy)] for p in pts]
    return np.array(xy), locx, locy, datumPt


def convert2DPointsTo3DCoordinateSystem(xy, e1, e2, origin):
    A = np.matrix([e1, e2])
    xyzPrime = np.matrix(xy) * A
    return np.squeeze(np.asarray(xyzPrime + origin))


def circleFit2DErr(allParams, xData, yData):
    pts = [np.linalg.norm([x - allParams[0], y - allParams[1]]) - allParams[2] for x, y in zip(xData, yData)]
    return (np.array(pts) ** 2).sum()


def fitCircle2D(pts, xGuess=0.0, yGuess=0.0, radiusGuess=1.0):
    X, Y = pts[:, 0], pts[:, 1]
    paramsToOptimize = np.array([xGuess, yGuess, radiusGuess])
    optimizedParams = optimize.fmin_powell(circleFit2DErr, paramsToOptimize, args=(X, Y))  # @UndefinedVariable
    xf = optimizedParams[0]
    yf = optimizedParams[1]
    rf = optimizedParams[2]
    return (xf, yf, rf)


def fitCircleRANSAC(x, y, tolerance, FRACTION_PTS_USE_FOR_CONVERGENCE,
                    MAX_ITERATIONS, EXCLUDE_INNERS=False, DEBUG=False):
    """

    :param x:
    :param y:
    :param tolerance:
    :param FRACTION_PTS_USE_FOR_CONVERGENCE:
    :param MAX_ITERATIONS:
    :param EXCLUDE_INNERS:
    :param DEBUG:
    :return:
    """
    count = 0
    nPts = len(x)
    Sbest = []
    while count < MAX_ITERATIONS:
        count += 1
        Si, Sx = [], []
        id0 = np.random.randint(nPts)
        id1 = np.random.randint(nPts)
        id2 = np.random.randint(nPts)
        if (id0 == id1) | (id1 == id2) | (id0 == id2):
            continue
        ma = (y[id1] - y[id0]) / (x[id1] - x[id0])
        mb = (y[id2] - y[id1]) / (x[id2] - x[id1])
        xc = (ma * mb * (y[id0] - y[id2]) + mb * (x[id0] + x[id1]) - ma * (x[id1] + x[id2])) / (2.0 * (mb - ma))
        yc = -1 * (1.0 / ma) * (xc - (x[id0] + x[id1]) / 2.0) + (y[id0] + y[id1]) / 2.0
        r = np.sqrt((xc - x[id0]) ** 2.0 + (yc - y[id0]) ** 2.0)
        for k1 in range(nPts):
            ri = np.sqrt((x[k1] - xc) ** 2.0 + (y[k1] - yc) ** 2.0)
            if EXCLUDE_INNERS:
                if ri < r:
                    continue
            dr = np.sqrt((ri - r) ** 2.0)
            if dr < tolerance:
                Si.append(k1)
            else:
                Sx.append(k1)
        if (len(Si) / nPts) > FRACTION_PTS_USE_FOR_CONVERGENCE:
            Sbest = Si
            xBest, yBest, rBest, countBest = xc, yc, r, count
            if DEBUG:
                print('Terminated on large Si')
            break
        if len(Si) > len(Sbest):
            Sbest = Si
            xBest, yBest, rBest, countBest = xc, yc, r, count
    xf, yf, rf = fitCircle2D(np.vstack((x[Sbest], y[Sbest])).T, xBest, yBest, rBest)
    return xf, yf, rf


def fitCircleRANSAC3D_xyz(xyz, tolerance, FRACTION_PTS_USE_FOR_CONVERGENCE,
                      MAX_ITERATIONS, EXCLUDE_INNERS=False, DEBUG=False):
    return fitCircleRANSAC3D(xyz[:,0],xyz[:,1],xyz[:,2], tolerance, FRACTION_PTS_USE_FOR_CONVERGENCE,
                      MAX_ITERATIONS, EXCLUDE_INNERS=False, DEBUG=False )


def fitCircleRANSAC3D(x, y, z, tolerance, FRACTION_PTS_USE_FOR_CONVERGENCE,
                      MAX_ITERATIONS, EXCLUDE_INNERS=False, DEBUG=False):
    """
    Rotate to plane of best fit, then 2D fit, rotate results back to 3D
        Return fitted point, R, norm
    :param x:
    :param y:
    :param z:
    :param tolerance:
    :param FRACTION_PTS_USE_FOR_CONVERGENCE:
    :param MAX_ITERATIONS:
    :param EXCLUDE_INNERS:
    :param DEBUG:
    :return:  xf, yf, zf, rf, normalVec
    """
    xy, e1, e2, origin = project3DPointsToPlanarCoordinateSystem(np.vstack((x, y, z)))
    xf, yf, rf = fitCircleRANSAC(xy[:, 0], xy[:, 1], tolerance, FRACTION_PTS_USE_FOR_CONVERGENCE,
                                 MAX_ITERATIONS, EXCLUDE_INNERS, DEBUG)
    xyzf = convert2DPointsTo3DCoordinateSystem([xf, yf], e1, e2, origin)
    return xyzf, rf, np.cross(e1, e2)


# ======================================================================================================================
#           GEOMETRY
# ======================================================================================================================
def buildCircle3D(X, N, R, nPts=50):
    """

    :param X: centre of circle
    :param N: normal of circle
    :param R: radius of circle
    :param nPts: number of points to place around circumference [50]
    :return: xyz of points on circumference of circle
    """
    ### Let W = (w0,w1,w2). Choose U = (u0,u1,u2) and V = (v0,v1,v2) using
    normalVec = N / np.linalg.norm(N)
    if (abs(normalVec[0]) >= abs(normalVec[1])):
        factor = 1.0 / np.sqrt(normalVec[0] * normalVec[0] + normalVec[2] * normalVec[2])
        u0 = -normalVec[2] * factor
        u1 = 0.0
        u2 = normalVec[0] * factor
    else:
        factor = 1.0 / np.sqrt(normalVec[1] * normalVec[1] + normalVec[2] * normalVec[2])
        u0 = 0.0
        u1 = normalVec[2] * factor
        u2 = -normalVec[1] * factor
    u = np.array([u0, u1, u2])
    v = np.cross(normalVec, u)
    ### X(t) = C + (r*cos(t))*U + (r*sin(t))*V ##
    theta = 0
    dtheta = 2 * np.pi / nPts
    fc_circle = np.zeros((3, nPts))
    for k in range(nPts):
        theta = theta + dtheta
        fc_circle[:, k] = (X + (R * np.cos(theta)) * u + (R * np.sin(theta)) * v).T

    return fc_circle.T


def fitPlaneToPointCloud_RANSAC(pts, planeSearchDist_abs, planeFractionToInclude, max_iterations=1000, random_state=None):
    """
    Fits a plane to a 3D point cloud using RANSAC.
        Choose random point
        Take next two closest points - calc plane
        Test number of points within planeSearchDist_abs 
            If > planeFractionToInclude
                test RMSE vs best

    Parameters:
    - pts: A numpy array of shape (N, 3) representing the 3D point cloud.
    - planeSearchDist_abs: The absolute distance to search for points for plane RMSE calc.
    - planePercentageToInclude: The fraction of points required to be part of plane.
    - max_iterations: The maximum number of iterations for the RANSAC algorithm (default=1000).
    - random_state: Random state for reproducibility (default=None).

    Returns:
    - plane_normal: A numpy array representing the normal vector of the fitted plane.
    - plane_center: A numpy array representing a point on the fitted plane.
    """
    best_plane_normal = None
    best_plane_center = None
    best_num_inliers = 0

    total_points = pts.shape[0]
    target_inliers = int(total_points * planeFractionToInclude)
    rndIDs = list(range(total_points))
    np.random.shuffle(rndIDs)
    for k1 in range(min(total_points, max_iterations)):
        # Randomly sample point - get closest 2 neighbours
        rndID = rndIDs[k1]
        idists = distPointPoints(pts[rndID], pts)
        sampleIDs = np.argsort(idists)[:3]
        samplePts = pts[sampleIDs]

        # Calculate plane parameters using these points
        v1 = samplePts[1] - samplePts[0]
        v2 = samplePts[2] - samplePts[0]
        plane_normal = np.cross(v1, v2)
        plane_normal /= np.linalg.norm(plane_normal)
        plane_center = np.mean(samplePts, axis=0)

        # Calculate distances from points to plane
        distances = np.abs(np.dot(pts - plane_center, plane_normal))
        


        # Count inliers
        inlier_mask = distances < planeSearchDist_abs
        num_inliers = np.sum(inlier_mask)
        if k1 > 0:
            if num_inliers > 5:
                # Add a penalty if all pts at boundary
                selectedPts = pts[inlier_mask,:]
                distFromCenter = min(distPointPoints(np.mean(selectedPts, axis=0), selectedPts))
                if distFromCenter > planeSearchDist_abs:
                    continue

        # Update best model if necessary
        if (best_plane_normal is None) or ((num_inliers >= target_inliers) and (num_inliers > best_num_inliers)):
            print(f"    Update best num_inliers = {num_inliers} ({num_inliers/total_points*100:0.2f}%)")
            best_plane_normal = plane_normal
            best_plane_center = plane_center
            best_num_inliers = num_inliers
    return best_plane_normal, best_plane_center


# ======================================================================================================================
# def findCylinderAxis(point_cloud):
#     """
#     Finds the main axis of a cylindrical point cloud.
    
#     :param point_cloud: Nx3 numpy array where N is the number of points and each row is (x, y, z).
#     :return: A unit vector defining the main axis of the cylinder.
#     """
#     from sklearn.decomposition import PCA
#     point_cloud_centered = point_cloud - np.mean(point_cloud, axis=0)
#     pca = PCA(n_components=3)
#     pca.fit(point_cloud_centered)
#     main_axis = pca.components_[0]
#     return main_axis / np.linalg.norm(main_axis)


# ======================================================================================================================
# ======================================================================================================================

def deg2rad(deg):
    return np.pi * deg / 180.0


def rad2deg(rad):
    return 180.0 * rad / np.pi


def rotateArray(velArray, rotationMatrix):
    vel = np.array(velArray)
    dims = vel.shape
    if dims[-1] != 3:
        raise AttributeError('ERROR ARRAY IS WRONG SIZE')
    pts = np.prod(dims[:-1])
    newDims = (pts, 3)
    vel = np.reshape(vel, newDims)
    velM = np.matrix(vel)
    velM = np.transpose(velM)
    velMR = np.zeros((pts, 3))
    for i in range(pts):
        velMR[i, :] = np.squeeze(rotationMatrix * velM[:, i])
    velOut = np.reshape(velMR, dims)
    return velOut


def rotationMatrixFromThreeAngles(xTheta, yTheta, zTheta):
    X = np.array([[1, 0, 0], [0, np.cos(xTheta), -np.sin(xTheta)], [0, np.sin(xTheta), np.cos(xTheta)]])
    Y = np.array([[np.cos(yTheta), 0, np.sin(yTheta)], [0, 1, 0], [-np.sin(yTheta), 0, np.cos(yTheta)]])
    Z = np.array([[np.cos(zTheta), -np.sin(zTheta), 0], [np.sin(zTheta), np.cos(zTheta), 0], [0, 0, 1]])
    return np.dot(np.dot(Z, Y), X)


def getClosestInSortedList(listIn, ref, INCREASEING=True):
    for k1 in range(0, len(listIn)):
        if INCREASEING:
            if listIn[k1] >= ref:
                return k1
        else:
            if listIn[k1] < ref:
                return k1
    return len(listIn) - 1


def getIdOfClosestFloat(listIn, ref):
    dd = np.array(listIn) - np.array(ref)
    return np.argmin(abs(dd))


def polar2Cart(r,theta):
    z = polar2z(r, theta)
    return np.real(z), np.imag(z)


def polar2z(r,theta):
    return r * np.exp( 1j * theta )


def z2Polar(z):
    return ( abs(z), np.angle(z) )


def cart2Polar(x,y):
    z = x + 1j * y
    return z2Polar(z)


def unit_vector(vector):
    """ Returns the unit vector of the vector.  """
    VM = np.linalg.norm(vector)
    if VM < 0.0000000000001:
        return np.array([0.0, 0.0, 0.0])
    return vector / VM


def vecFromPtToLine(pt, ptOnLine, lineNorm):
    return (ptOnLine - pt) - (np.dot(ptOnLine - pt, lineNorm) * lineNorm)


def fcdot(u, v):
    """ dot product of two lists of vecs"""
    u, v = __forcePts_nx3(u), __forcePts_nx3(v)
    return np.array([np.dot(i, j) for i, j in zip(u, v)])


def cosineDifferenceBetweenTwoMats(mA, mB):
    dims = mA.shape
    matDiff = np.zeros(dims[:-1])
    for k in range(dims[2]):
        for j in range(dims[1]):
            for i in range(dims[0]):
                matDiff[i, j, k] = angleBetween2Vec(mA[i, j, k], mB[i, j, k])

    return matDiff


def getVectorComponentAlongRefVec(velVec, refVec):
    """ Returns vec parallel to refVec"""
    refVec_U = refVec / np.linalg.norm(refVec)
    return np.dot(velVec, refVec_U) * refVec_U


def getVectorComponentAlongRefVec_Mag(velVec, refVec):
    """ Returns magnitude (signed) of vec parallel to refVec"""
    refVec_U = refVec / np.linalg.norm(refVec)
    llv = np.dot(velVec, refVec_U) * refVec_U
    llvM = vectorMagnitudes(llv)
    if areVecsConsistent(llv, refVec):
        return  llvM
    else:
        return  -1*llvM


def getVectorComponentNormalToRefVec(velVec, refVec):
    """ Returns vec component in plane normal to refVec"""
    vel_t = getVectorComponentAlongRefVec(velVec, refVec)
    return velVec - vel_t


def closeCurve(xyz):
    pts = np.array(xyz)
    pts = np.vstack((pts, pts[0]))
    return pts


def splinePoints(xyz, nSplinePts, periodic=0, RETURN_NUMPY=False, smooth=0, u=None, weights=None, derivitive=0):
    """
     returns list, shape (3, n), unless ask for numpy then transpose
    """
    pts = np.asarray(xyz)

    if periodic>0:
        smooth = 0.0
        pts = np.vstack((pts, pts[0]))
        if weights is not None:
            weights = np.hstack((weights, weights[-1]))
    pts = pts.transpose()
    tck, u = interpolate.splprep(pts, u=u, s=smooth, per=periodic, w=weights)
    uu = np.linspace(u.min(), u.max(), nSplinePts)
    newPts = interpolate.splev(uu, tck, der=derivitive)
    if RETURN_NUMPY:
        return np.array(newPts).T
    return newPts


def splinePoints_RansacSmooth(xyz, periodic=0, RETURN_NUMPY=False):
    """
     returns list, shape (3, n), unless ask for numpy then transpose
    """
    pts = np.asarray(xyz)

    if periodic>0:
        smooth = 0.0
        pts = np.vstack((pts, pts[0]))
    pts = pts.transpose()
    tck, u = interpolate.splprep(pts, per=periodic)
    uu = np.linspace(u.min(), u.max(), len(xyz))
    newPts = interpolate.splev(uu, tck, der=1)
    print(np.sum(np.sqrt(np.power(newPts, 2.0))))



def splineFunction(x, y, kind):
    return interpolate.interp1d(x, y, bounds_error=False, kind=kind)


def splineXY(x, y, nPts, kind='linear', per=False):
    """ Input : x, y, numPts
        Output : X, Y
    """
    if per:
        xp = np.hstack((np.array(x), x[-1]+(x[1]-x[0])))
        y = np.hstack((np.array(y), y[0]))
        f = splineFunction(xp, y, kind)
    else:
        f = splineFunction(x, y, kind)
    xx = np.linspace(min(x), max(x), nPts)
    yy = f(xx)
    return xx, yy


def splineCurve(x, y, nPts, interpolation_method='cubic'):
    """Spline is 2D (i.e x not always increasing)

    Args:
        x (np.array): x array
        y (np.array): y array
        interpolation_methods (str, optional): of 'slinear', 'quadratic', 'cubic']. Defaults to 'cubic'.
    """
    points = np.vstack([x,y]).T
    distance = np.cumsum( np.sqrt(np.sum( np.diff(points, axis=0)**2, axis=1 )) )
    distance = np.insert(distance, 0, 0)/distance[-1]
    alpha = np.linspace(0, 1, nPts)
    interpolator = interpolate.interp1d(distance, points, bounds_error=False, kind=interpolation_method, axis=0)
    res = interpolator(alpha)
    return res.T


def getMaxsMinsInflections(X):
    """
    Simple extrema finder
    :param X: 1D array
    :return: linst of maxima, list of minima, lint of inflections
    """
    from scipy.signal import argrelextrema
    gg = np.gradient(X)
    return argrelextrema(X, np.greater)[0], \
           argrelextrema(X, np.less)[0], \
           np.hstack((argrelextrema(gg, np.greater)[0],argrelextrema(gg, np.less)[0]))


def sigmoidFunction(value, center=0.0, gradient=1.0):
    """
    calculate sigmoid function of input value (np array)
    center is what to adjust input by so zero centered
    gradient is used to adjust slope of sigmoid -
        (adjusted) values above this number will have output > 0.72
    """
    valueT = value - center
    output = 1.0 / (1.0 + np.exp(-valueT / gradient))
    return output


def extractContinuousSegFromPeriodic(xx, IDstart, IDend, TIME=False):
    """
    Take a sub section from an array, assuming periodic (to deal with subsection going beyond bounds)
    """
    N = len(xx)
    if TIME:
        delta = np.mean(np.diff(xx))
        tN = xx[-1]
        xx3 = np.hstack((xx, xx+delta+tN, xx+(2.0*delta)+(2.0*tN)))
    else:
        xx3 = np.hstack((xx,xx,xx))
    IDstart = IDstart + N
    IDend = IDend + N
    xxOut = xx3[IDstart:IDend]
    if TIME:
        return xxOut - tN - delta
    return xxOut


def fitGaussianToData(xx, yy):
    popt, _ = optimize.curve_fit(gaussianFunction, xx, yy)
    center, width, height = popt
    # 'x': xSub, 'y': ftk.gaussianFunction(xSub, *popt)
    return center, width, height


def gaussianFunction(value, center=0.0, width=1.0, height=1.0):
    """
    calculate gaussian function of input value (np array)
    center is what to adjust input by so zero centered
    width is standard dev of bell curve
    height is height of peak
    """
    valueT = value - center
    output = height * (np.exp(-(valueT ** 2.0) / (2.0 * width ** 2.0)))
    return output


def calculate_signed_area_2d(points_2d):
    # Use the shoelace formula to calculate the signed area of the 2D points
    n = len(points_2d)
    area = 0.0
    for i in range(n):
        x1, y1 = points_2d[i]
        x2, y2 = points_2d[(i + 1) % n]
        area += (x1 * y2 - y1 * x2)
    return area / 2.0


def isClosedPolygonClockwise(xyzIn, refVec):
    # Calculate the centroid to use as origin
    xyP = project3DPointsToPlanarCoordinateSystem(xyzIn, refVec)[0]
    # Calculate the signed area
    signed_area = calculate_signed_area_2d(xyP)
    # If the signed area is negative, it's clockwise; otherwise, it's counterclockwise
    return signed_area < 0


def ensureClosedPolyIsClockwise(xzyIn, refVec):
    if isClosedPolygonClockwise(xzyIn, refVec):
        return xzyIn
    return xzyIn[::-1]

def setFirstPtOfPolygon(xyzIn, refVec):
    """
    Reorder the points of a polygon so that the first point is closest to the refVec (e.g. '12 o'clock' position).

    This function assumes the polygon is planar and reorders the points so that the first point
    is the one closest to the end of a vector starting from the polygon's center point
    and pointing in the direction of the reference vector.

    Parameters:
    xyzIn (array-like): Input points of the polygon.
    refVec (array-like): Reference vector indicating the '12 o'clock' direction.

    Returns:
    numpy.ndarray: Reordered points of the polygon.
    """
    pts = np.array(xyzIn)
    refVec = np.array(refVec)
    nPts = len(pts)
    cp = getPolygonCenterPoint(xyzIn)
    meanDiameter = getPolygonMeanDiameter(xyzIn)
    vectorEnd = cp + meanDiameter / 1.0 * refVec
    closestPtIndex = 0
    ptToLineDist = distTwoPoints(vectorEnd, pts[0])
    for i in range(1, nPts, 1):
        iDist = distTwoPoints(vectorEnd, pts[i])
        if (iDist < ptToLineDist):
            closestPtIndex, ptToLineDist = i, iDist
    return reorderPointsStatingAti(pts, closestPtIndex)


def reorderPointsStartClosestToX(pts, X):
    ID = getIDOfClosestPoint(X, pts)
    return reorderPointsStatingAti(pts, ID)


def reorderPointsStatingAti(pts, i):
    pts = __forcePts_nx3(pts)
    ptsOut = np.vstack((pts[i:][:], pts[:i][:]))
    return ptsOut


def sortPointsByClosest_RETURNIDs(xyz, refX0):
    """ Grab closest to refX0, grab next closest, keep going in that direction
     RETURN IDs """
    X = [i for i in xyz]
    X2 = [i for i in xyz]
    IDs = [getIDOfClosestPoint(refX0, X)]
    X2.pop(IDs[0])
    while True:
        try:
            nextID = getIdOfPointClosestToX(X[IDs[-1]], X2)
            IDs.append(X.index(X2[nextID]))
            X2.pop(nextID)
        except ValueError:
            break
    return IDs


def sortPointsByClosest(xyz):
    """ Grab first, grab closest, keep going in that direction """
    X = [i for i in xyz]
    ptsOut = [X.pop(0)]
    while True:
        try:
            nextID = getIdOfPointClosestToX(ptsOut[-1], X)
            ptsOut.append(X.pop(nextID))
        except ValueError:
            break
    return np.array(ptsOut)


def cumulativeDistanceAlongLine(xyzPts):
    """ Return array same len as input with sum of dists
        [0.0, dist-0to1, dist-0to1to2, ...] """
    if len(xyzPts[0, :]) == 2:
        dists = [np.sqrt((p1[0] - p2[0]) ** 2 + (p1[1] - p2[1]) ** 2) for
                 p2, p1 in zip(xyzPts[:-1], xyzPts[1:])]
    else:
        xyzPts = __forcePts_nx3(xyzPts)
        dists = [np.sqrt((p1[0] - p2[0]) ** 2 + (p1[1] - p2[1]) ** 2 + (p1[2] - p2[2]) ** 2) for
                 p2, p1 in zip(xyzPts[:-1], xyzPts[1:])]
    dists.insert(0, 0.0)
    return np.cumsum(dists)


def getIdOfPointClosestToX(X, xyz):
    diffs = distPointPoints(X, xyz)
    return np.argmin(diffs)


def getPolygonMeanRadius(xyzIn, EXCLUDE_CENTER=True):
    pts = np.array(xyzIn)
    nPts = len(pts)
    cp = np.mean(pts, 0)
    allR = []
    for i in range(nPts):
        allR.append(distPointPoints(cp, pts[i]))
    if EXCLUDE_CENTER:
        allR = sorted(allR)
        return np.mean(allR[1:])
    else:
        return np.mean(allR)


def getPolygonMeanDiameter(xyzIn):
    return getPolygonMeanRadius(xyzIn) * 2.0


def getPolygonCenterPoint(xyzIn):
    pts = np.array(xyzIn)
    return np.mean(pts, 0)


def lineMagnitude(lineStartxyz, lineEndxyz):
    if len(lineStartxyz) == 3:
        lineMagnitude = np.sqrt(np.power((lineEndxyz[0] - lineStartxyz[0]), 2) +
                                np.power((lineEndxyz[1] - lineStartxyz[1]), 2) +
                                np.power((lineEndxyz[2] - lineStartxyz[2]), 2))
    else:
        lineMagnitude = np.sqrt(np.power((lineEndxyz[0] - lineStartxyz[0]), 2) +
                                np.power((lineEndxyz[1] - lineStartxyz[1]), 2))
    return lineMagnitude


def distancePointToLineSegPerpendicular(lineStartxyz, lineEndxyz, ptxyz):
    return np.linalg.norm(np.cross(lineEndxyz-lineStartxyz, lineStartxyz-ptxyz))/np.linalg.norm(lineEndxyz-lineStartxyz)


def ndarrayToListOfTuple3(ndArrayIn):
    xyzOut = []
    for ndarray3 in ndArrayIn:
        xyzOut.append(tuple(ndarray3))

    return xyzOut


def diff(myList):
    a = np.array(myList[:-1])
    b = np.array(myList[1:])
    mydiff = b - a
    return mydiff


def circleFit2DErr(allParams, xData, yData):
    """ Fit a circle to points
    """
    # pts = [(np.linalg.norm([x-allParams[0],y-allParams[1]])-allParams[2])*w for x,y,w in zip(xData,yData,allParams[3:])]
    pts = [np.linalg.norm([x - allParams[0], y - allParams[1]]) - allParams[2] for x, y in zip(xData, yData)]
    return (np.array(pts) ** 2).sum()

def distanceToPlane(p0, n, x):
    # Plane through point p0, with normal n. Point x
    return np.dot(np.array(n), np.array(x) - np.array(p0))

def residualPlane(parameters, X, Y, Z):
    px, py, pz, theta, phi = parameters
    nx, ny, nz = np.sin(theta) * np.cos(phi), np.sin(theta) * np.sin(phi), np.cos(theta)
    distances = [distanceToPlane([px, py, pz], [nx, ny, nz], [x, y, z]) for x, y, z in zip(X, Y, Z)]
    return distances

def residualsCircle(parameters, sArr, rArr, point, X, Y, Z):
    r, s, Ri = parameters
    planePointArr = s * sArr + r * rArr + np.array(point)
    distance = [np.linalg.norm(planePointArr - np.array([x, y, z])) for x, y, z in zip(X, Y, Z)]
    res = [(Ri - dist) for dist in distance]
    return res


def icp(sourcePts, targetPts, max_iterations=100, tolerance=1e-6):
    """
    Iterative Closest Point (ICP) algorithm for point set registration.

    Parameters:
    - X: A numpy array of shape (N, 3) representing the source set of 3D points.
    - Y: A numpy array of shape (M, 3) representing the target set of 3D points.
    - max_iterations: Maximum number of iterations for the algorithm (default=100).
    - tolerance: Convergence criterion based on the change in error (default=1e-6).

    Returns:
    - R_best: A 3x3 numpy array representing the best rotation matrix.
    - t_best: A 1x3 numpy array representing the best translation vector.
    """
    R_best = np.eye(3)  # Initial best rotation matrix (identity)
    t_best = np.zeros(3)  # Initial best translation vector
    min_error = np.inf  # Initial minimum error
    sourcePtsBest = None
    rndIt_every = int(max_iterations/5.0)
    tree = cKDTree(targetPts)
    for k1 in range(max_iterations):
        distances, indices = tree.query(sourcePts)
        Y_matched = targetPts[indices]
        centroid_X = np.mean(sourcePts, axis=0)
        centroid_Y = np.mean(Y_matched, axis=0)

        X_centered = sourcePts - centroid_X
        Y_centered = Y_matched - centroid_Y

        covariance_matrix = np.dot(Y_centered.T, X_centered)
        U, _, Vt = np.linalg.svd(covariance_matrix)

        R_new = np.dot(U, Vt)
        t_new = centroid_Y - np.dot(R_new, centroid_X)

        sourcePts = np.dot(sourcePts, R_new.T) + t_new

        total_error = np.sum(distances)

        if total_error < min_error:
            min_error = total_error
            R_best = np.dot(R_new, R_best)
            t_best = np.dot(R_new, t_best) + t_new
            sourcePtsBest = sourcePts

        if total_error < tolerance:
            break

    return R_best, t_best, sourcePtsBest

def fitPlane3DOptimize(pts):
    x,y,z = pts[:,0], pts[:,1], pts[:,2]
    estPlane = np.array([np.mean(x), np.mean(y), np.mean(z), 0, 0])
    bestFitValues = optimize.leastsq(residualPlane, estPlane, args=(x, y, z))[0]
    xF, yF, zF, tF, pF = bestFitValues
    point = np.array([xF, yF, zF])
    normal = np.array([np.sin(tF) * np.cos(pF), np.sin(tF) * np.sin(pF), np.cos(tF)])
    return normal, point


def fitCircle3D(x, y, z, xm, ym, zm, rm):
    """
    Fit a circle to 3D points.

    Parameters:
    - x, y, z: Arrays of x, y, and z coordinates of the points.
    - xm, ym, zm: Coordinates of the center of the circle.
    - rm: Radius of the circle.

    Returns:
    - centerPointArr: Coordinates of the center of the circle.
    - RiF: Radius of the circle.
    - normal: Normal vector of the plane containing the circle.
    """
    estPlane = np.array([xm, ym, zm, 0, 0])
    # print residualsPlane(estPlane, pts)
    bestFitValues = optimize.leastsq(residualPlane, estPlane, args=(x, y, z))[0]
    xF, yF, zF, tF, pF = bestFitValues

    point = [xF, yF, zF]
    normal = [np.sin(tF) * np.cos(pF), np.sin(tF) * np.sin(pF), np.cos(tF)]

    # Fitting a circle inside the plane
    # creating two inplane vectors
    sArr = np.cross(np.array([1, 0, 0]), np.array(normal))  # assuming that normal not parallel x!
    sArr = sArr / np.linalg.norm(sArr)
    rArr = np.cross(sArr, np.array(normal))
    rArr = rArr / np.linalg.norm(rArr)  # should be normalized already, but anyhow

    estimateCircle = np.array([0.0, 0.0, rm])  # px,py,pz and zeta, phi
    bestCircleFitValues = optimize.leastsq(residualsCircle, estimateCircle, args=(sArr, rArr, point, x, y, z))[0]

    rF, sF, RiF = bestCircleFitValues
    centerPointArr = sF * sArr + rF * rArr + np.array(point)
    # synthetic=[list(centerPointArr+ RiF*np.cos(phi)*rArr+RiF*np.sin(phi)*sArr) for phi in np.linspace(0, 2*np.pi,50)]
    # [cxTupel,cyTupel,czTupel]=[ x for x in zip(*synthetic)]
    return (centerPointArr, RiF, np.cross(sArr, rArr))


def lagBetweenTwoCurves(y1: np.array, y2: np.array, timeArray: np.array = None):
    """
    Calculate the lag (delay) between two similar time series curves.

    This function uses cross-correlation to determine the time delay between two signals.

    Parameters:
    - y1 (np.array): First time series.
    - y2 (np.array): Second time series.
    - timeArray (np.array, optional): Array of time points corresponding to y1 and y2. 
      If None, a normalized time array from -0.5 to 0.5 is used. Default is None.

    Returns:
    - float: The time delay between the two signals. A positive value indicates 
             that y2 lags behind y1, while a negative value indicates that y1 lags behind y2.

    Note:
    - The two input arrays (y1 and y2) must have the same length.
    - If timeArray is provided, it must have the same length as y1 and y2.
    """
    n = len(y1)
    corr = np.correlate(y2, y1, mode='same') / np.sqrt(np.correlate(y1, y1, mode='same')[int(n/2)] * np.correlate(y2, y2, mode='same')[int(n/2)])
    if timeArray is None:
        delay_arr = np.linspace(-0.5, 0.5, n)
    else:
        delay_arr = np.roll(timeArray, int(n/2))
    delay = delay_arr[np.argmax(corr)]
    return delay


def calculatePlaneThreePoints(p0, p1, p2):
    """ Plane from three points (v0(p0 to p1) and v1(p0 to p2) -> cross == norm, p2 for D)
        Return [A,B,C,D]
    """
    v0, v1 = np.array(p1) - np.array(p0), np.array(p2) - np.array(p0)
    n0 = np.cross(v0, v1)
    n0 = n0 / np.linalg.norm(n0)
    return getPlaneConstantsFromPointAndVector(p2, n0)


def getPlaneConstantsFromPointAndVector(pt, vector):
    D = -pt[0] * vector[0] - pt[1] * vector[1] - pt[2] * vector[2]
    plane = np.hstack((vector, D))
    return plane


def calculateDerivativeOnLine_backwardDiff(pim1, pi, dt):
    """ Calculate the derivative of a line using backward difference. 
    """
    dr_dt = [(m - n) / dt for m, n in zip(pim1, pi)]
    return dr_dt


def calculateSecondDerivativeOnLine(pim1, pi, pip1, dt):
    d2r_dt2 = [(p - 2.0 * n + m) / dt ** 2 for m, n, p in zip(pim1, pi, pip1)]
    return d2r_dt2


def __calcTNB(dr_dt, d2r_dt2):
    T = dr_dt / np.linalg.norm(dr_dt)
    N = np.cross(dr_dt, np.cross(d2r_dt2, dr_dt)) / \
        (np.linalg.norm(dr_dt) * np.linalg.norm(np.cross(d2r_dt2, dr_dt)))
    B = np.cross(T, N)
    return (T, N, B)


def calculateCurvature3Pts(pim1, pi, pip1, dx):
    dr_dt = calculateDerivativeOnLine_backwardDiff(pim1, pi, dx)
    d2r_dt2 = calculateSecondDerivativeOnLine(pim1, pi, pip1, dx)
    num = np.linalg.norm(np.cross(dr_dt, d2r_dt2))
    denom = np.power(np.linalg.norm(dr_dt), 3)
    k = num / denom
    TNB = __calcTNB(dr_dt, d2r_dt2)
    return (k, TNB)


def calculateCurvature_D(gf: np.array, ggf: np.array):
    num = [np.linalg.norm(np.cross(i, j)) for i, j in zip(gf, ggf)]
    denom = [np.power(np.linalg.norm(i), 3) for i in gf]
    k = np.array(num) / np.array(denom)
    TNB = [__calcTNB(i, j) for i, j in zip(gf, ggf)]
    return (k, TNB)


def setVectorDirection(vecIn, oo, pointMorePositive):
    """ Will ensure vector at point oo is pointing towards 'pointMorePositive' else flip"""
    oo, vecIn = np.asarray(oo), np.asarray(vecIn)
    dp, dm = distTwoPoints(oo + vecIn, pointMorePositive), distTwoPoints(oo - vecIn, pointMorePositive)
    if dp > dm:
        return -1 * vecIn
    return vecIn


def buildRotationMatrix(vec, theta_rad):
    vec = vec / np.linalg.norm(vec)
    RotM = [[np.cos(theta_rad) + vec[0] ** 2 * (1.0 - np.cos(theta_rad)), \
             vec[0] * vec[1] * (1.0 - np.cos(theta_rad)) - vec[2] * np.sin(theta_rad), \
             vec[1] * np.sin(theta_rad) + vec[0] * vec[2] * (1.0 - np.cos(theta_rad))], \
            [vec[2] * np.sin(theta_rad) + vec[0] * vec[1] * (1.0 - np.cos(theta_rad)), \
             np.cos(theta_rad) + vec[1] ** 2 * (1.0 - np.cos(theta_rad)), \
             -vec[0] * np.sin(theta_rad) + vec[1] * vec[2] * (1.0 - np.cos(theta_rad))], \
            [-vec[1] * np.sin(theta_rad) + vec[0] * vec[2] * (1.0 - np.cos(theta_rad)), \
             vec[0] * np.sin(theta_rad) + vec[1] * vec[2] * (1.0 - np.cos(theta_rad)), \
             np.cos(theta_rad) + vec[2] ** 2 * (1.0 - np.cos(theta_rad))]]
    return RotM


def adjustPointsCenterOfMassToOrigin(pts):
    pts = __forcePts_nx3(pts)
    centerOfMass = np.mean(pts, 0)
    return pts - centerOfMass


def areVecsConsistent(vecA, vecB):
    vecA = np.array(vecA)
    vecB = np.array(vecB)
    nA = np.linalg.norm(vecA)
    nB = np.linalg.norm(vecB)
    if nA > 0:
        vecA = vecA / nA
    if nB > 0:
        vecB = vecB / nB
    if np.linalg.norm(vecB - vecA) > np.linalg.norm(vecB - -1.0 * vecA):
        return False
    return True


def setVecAConsitentWithVecB(vecA, vecB):
    """
     Make vec dir consistant with poly order
    """
    if not areVecsConsistent(vecA, vecB):
        vecA = -1.0 * vecA
    return vecA


def buildConvexHUll3D(xyz, nPts=25, planeABCD=None, TO_SPLINE=True):
    try:
        len(planeABCD)
    except TypeError:
        planeABCD = fitPlaneToPoints(xyz)
    xy, e1, e2, origin = project3DPointsToPlanarCoordinateSystem(xyz, planeABCD[:3])
    xyL = [tuple(i) for i in xy]
    chxy = convexHull(xyL)
    if TO_SPLINE:
        chxys = np.array(splinePoints(chxy, nPts, periodic=1)).T
        # chxys = splinePoints(np.array(chxy).T, nPts, periodic=1, RETURN_NUMPY=True)
    else:
        chxys = np.array(chxy)
    chxyz = convert2DPointsTo3DCoordinateSystem(chxys, e1, e2, origin)
    return np.asarray(chxyz)


def orderPointsConvexHull_IDs(xyz):
    planeABCD = fitPlaneToPoints(xyz)
    xy, e1, e2, origin = project3DPointsToPlanarCoordinateSystem(xyz, planeABCD[:3])
    xyL = [tuple(i) for i in xy]
    chxy = convexHull(xyL)
    IDs = []
    for k1 in range(len(chxy)):
        dd = distPointPoints(chxy[k1], xy)
        iID = np.argmin(dd)
        IDs.append(iID)
    return IDs


def orderPointsConvexHull(xyz):
    IDs = orderPointsConvexHull_IDs(xyz)
    pts_out = np.array([xyz[i] for i in IDs])
    return pts_out


def groupContinuousLinesByTol(ptsIn, tol):
    """ pts is N x 3 array of pts forming continuous lines (at least one)
        separated by gaps of 'tol'
        Returns list of pts separated
    """
    pts = ptsIn.copy()
    listOut = []
    pi = pts[0, :]
    currPts = [pi]
    pts = np.delete(pts, 0, 0)
    while len(pts) > 0:
        dd = distPointPoints(pi, pts)
        nextId = np.argmin(dd)
        if dd[nextId] < tol:
            pi = pts[nextId, :]
            currPts.append(pi)
            pts = np.delete(pts, nextId, 0)
        else:
            listOut.append(np.array(currPts))
            pi = pts[nextId, :]
            currPts = [pi]
            pts = np.delete(pts, nextId, 0)
    listOut.append(np.array(currPts))
    return listOut


def buildContinuousLineByClosestPt(ptsIn):
    """ pts is N x 3 array of pts
        returns points ordered by proximity
    """
    pts = ptsIn.copy()
    pi = pts[0, :]
    currPts = [pi]
    pts = np.delete(pts, 0, 0)
    while len(pts) > 0:
        dd = distPointPoints(pi, pts)
        nextId = np.argmin(dd)
        pi = pts[nextId, :]
        currPts.append(pi)
        pts = np.delete(pts, nextId, 0)
    return np.array(currPts)


def accumulatedAverage(arrayIn):
    """
        Calc array that is average from start to i of input
    """
    aOut = [np.mean(arrayIn[:i + 1]) for i in range(len(arrayIn))]
    return np.array(aOut)


def interpolateNANsFromSurroundingValues_linear(array):
    """ This is linear temporal interpolation
    """
    mask = np.isnan(array)
    array[mask] = np.interp(np.flatnonzero(mask), np.flatnonzero(~mask), array[~mask])
    return array


def interpolateNANsFromSurroundingValues_cubic(nparray):
    arrayTF = np.isnan(nparray)
    if sum(arrayTF) == 0:
        return nparray
    dims = nparray.shape
    gridX, gridY = np.mgrid[0:dims[0], 0:dims[1]]
    pp = [[i, j] for i, j in zip(gridX.flatten(), gridY.flatten()) if not arrayTF[i, j]]
    values = [nparray[i, j] for i, j in zip(gridX.flatten(), gridY.flatten()) if not arrayTF[i, j]]
    return interpolate.griddata(pp, values, (gridX, gridY), method='cubic')


# ============ LINE SEG PIERCE TRIANGLE ========================================
def doesLineSegPierceTriangle(lineP0, lineP1, triP0, triP1, triP2):
    if doesLinePierceTriangle(lineP0, lineP1, triP0, triP1, triP2):
        s4 = pluckerSideOperator([triP1, triP2], [triP0, lineP1])
        s5 = pluckerSideOperator([triP1, triP2], [lineP0, triP0])

        if (s4 < 0) & (s5 < 0):
            return True
        elif (s4 > 0) & (s5 > 0):
            return True
    return False


def doesLinePierceTriangle(lineP0, lineP1, triP0, triP1, triP2):
    s1 = pluckerSideOperator([lineP0, lineP1], [triP0, triP1])
    s3 = pluckerSideOperator([lineP0, lineP1], [triP1, triP2])
    s2 = pluckerSideOperator([lineP0, lineP1], [triP2, triP0])
    if (s1 < 0) & (s2 < 0) & (s3 < 0):
        return True
    elif (s1 > 0) & (s2 > 0) & (s3 > 0):
        return True

    return False


def pluckerSideOperator(line0, line1):
    a = pluckerLine(line0)
    b = pluckerLine(line1)
    value = a[0] * b[4] + a[1] * b[5] + a[2] * b[3] + a[3] * b[2] + a[4] * b[0] + a[5] * b[1]
    return value

def pluckerLine(line):
    # line is 2 x 3
    L0 = line[0][0] * line[1][1] - line[1][0] * line[0][1]
    L1 = line[0][0] * line[1][2] - line[1][0] * line[0][2]
    L2 = line[0][0] - line[1][0]
    L3 = line[0][1] * line[1][2] - line[1][1] * line[0][2]
    L4 = line[0][2] - line[1][2]
    L5 = line[1][1] - line[0][1]
    return [L0, L1, L2, L3, L4, L5]


def doesVectorPierceAnyTriangle(vS, vE, triPolyData):
    nTris = triPolyData.GetNumberOfCells()
    for k2 in range(nTris):
        t0 = triPolyData.GetPoint(triPolyData.GetCell(k2).GetPointIds().GetId(0))
        t1 = triPolyData.GetPoint(triPolyData.GetCell(k2).GetPointIds().GetId(1))
        t2 = triPolyData.GetPoint(triPolyData.GetCell(k2).GetPointIds().GetId(2))

        if doesLineSegPierceTriangle(vS, vE, t0, t1, t2):
            return k2
    return None


# ==========================================================================
# ==========================================================================
def calculateBSA(height_cm, weight_kg, METHOD='DuBois'):
    if METHOD == 'DuBois':
        BSA = 0.007184 * np.power(height_cm, 0.725) * np.power(weight_kg, 0.425)
    else:
        raise ValueError('Method unknown. Use one of: {DuBois}')
    return BSA


# ==========================================================================
# RANDOM CLASSES
# ==========================================================================
def interpolateNANsFromSurroundingValues_linear(array):
    """ This is linear temporal interpolation
    """
    mask = np.isnan(array)
    array[mask] = np.interp(np.flatnonzero(mask), np.flatnonzero(~mask), array[~mask])
    return array


######################################################################
# CONVEX HULL
######################################################################
"""
Calculate the convex hull of a set of n 2D-points in O(n log n) time.  
Taken from Berg et al., Computational Geometry, Springer-Verlag, 1997.
Dinu C. Gherman
"""
def _myDet(p, q, r):
    """Calc. determinant of a special matrix with three 2D points.
    The sign, "-" or "+", determines the side, right or left,
    respectivly, on which the point r lies, when measured against
    a directed vector from p to q.
    """
    # We use Sarrus' Rule to calculate the determinant.
    # (could also use the Numeric package...)
    sum1 = q[0] * r[1] + p[0] * q[1] + r[0] * p[1]
    sum2 = q[0] * p[1] + r[0] * q[1] + p[0] * r[1]
    return sum1 - sum2


def _isRightTurn(pqr):
    """Do the vectors pq:qr form a right turn, or not?"""
    (p, q, r) = pqr
    assert p != q and q != r and p != r
    if _myDet(p, q, r) < 0:
        return 1
    else:
        return 0


def _isPointInPolygon(r, P):
    """Is point r inside a given polygon P?"""
    # We assume the polygon is a list of points, listed clockwise!
    for i in range(len(P[:-1])):
        p, q = P[i], P[i + 1]
        if not _isRightTurn((p, q, r)):
            return 0  # Out!
    return 1  # It's within!


def convexHull(P):
    """Calculate the convex hull of a set of points."""
    # Remove any duplicates
    # If the hull has a duplicate point, it will be returned once
    # It is up to the application to handle it correctly
    unique = {}
    for p in P:
        unique[p] = 1
    points = list(unique.keys())
    points.sort()
    # Build upper half of the hull.
    upper = [points[0], points[1]]
    for p in points[2:]:
        upper.append(p)
        while len(upper) > 2 and not _isRightTurn(upper[-3:]):
            del upper[-2]
    # Build lower half of the hull.
    points.reverse()
    lower = [points[0], points[1]]
    for p in points[2:]:
        lower.append(p)
        while len(lower) > 2 and not _isRightTurn(lower[-3:]):
            del lower[-2]
    # Remove duplicates.
    del lower[0]
    del lower[-1]
    # Concatenate both halfs and return.
    return tuple(upper + lower)


######################################################################
######################################################################
